#-----------------------------------
# Imports
#-----------------------------------

# General Imports
import ctypes
from typing import Any, overload, Union

# PyQt5 Imports
from PyQt5.QtCore import QLineF, QPointF, Qt, QRectF
from PyQt5.QtGui import QFontMetricsF
from PyQt5.QtWidgets import QGraphicsItem, QGraphicsSceneDragDropEvent, QGraphicsSceneHoverEvent

# Project Imports
from muphyn.packages.core.plci_core_data_type import DataType
from muphyn.packages.core.utils.log_manager import LogManager
from muphyn.packages.interface.graphical_actions.diagram_link_nodes_action import DiagramLinkNodesAction
from muphyn.packages.interface.models.graphical_models.abstract_graphical_element import AbstractGraphicalElement
from muphyn.packages.interface.models.signals_model.signal_node_model import SignalNodeModel
from muphyn.packages.interface.models.signals_model.abstract_connection_model import AbstractConnectionModel
from muphyn.packages.interface.shapes.basic_shape.grouped_shapes import GroupedShapes
from muphyn.packages.interface.shapes.basic_shape.shapes import Line, Text, Polygon
from muphyn.packages.interface.utils.interface_constants import MuphynFonts

#-----------------------------------
# Class
#-----------------------------------

class InputConnectionModel (SignalNodeModel) :
    """Est la classe décrivant le fonctionnement des entrées des boxes."""
    # -------------
    # Static Values
    # -------------
    ConnectorPoints = [
        QPointF(0, AbstractConnectionModel.ConnectorSide - 3),
        QPointF(0, 2 * AbstractConnectionModel.ConnectorSide - 3),
        QPointF(AbstractConnectionModel.ConnectorSide, 3 * AbstractConnectionModel.ConnectorSide/2 - 3),
    ]
    LinePoints = [
        QPointF(AbstractConnectionModel.ConnectorSide, 3*AbstractConnectionModel.ConnectorSide/2 - 3),
        QPointF(AbstractConnectionModel.ConnectorSide + AbstractConnectionModel.ConnectorLineLength, 3*AbstractConnectionModel.ConnectorSide/2 - 3)
    ]

    # -------------
    # Constructors
    # -------------

    def __init__ (self, name : str, data_type : DataType, position : QPointF, text : str = '', is_infinite : bool = False, 
                  parent : QGraphicsItem = None) :

        SignalNodeModel.__init__(self, name, data_type, position, AbstractConnectionModel.ItemSize, [], text, is_infinite, parent)
        
        self.setAcceptDrops(True)
        
        # Connector Group
        self.connector_group = GroupedShapes(parent=self)
        self.connector_group.setAcceptHoverEvents(True)
        self.connector_group.hoverEnterEvent = self.connectorHoverEnterEvent
        self.connector_group.hoverLeaveEvent = self.connectorHoverLeaveEvent

        # Connector
        self._connector = Polygon(
            InputConnectionModel.ConnectorPoints,
            border_color=self._color,
            fill_color=Qt.GlobalColor.white,
            pen_join_style=Qt.PenJoinStyle.RoundJoin, 
            parent=self.connector_group
        )
        # Get Connector center
        self.connector_center = self._connector.boundingRect().center()
        
        # Line
        self._line = Line(QLineF(*InputConnectionModel.LinePoints), parent=self.connector_group)
        
        # Label
        self._font_metrics = QFontMetricsF(MuphynFonts.IOConnectorFont)
        self._lbl = Text(
            name, QPointF(-3, 0), 
            alignment=Qt.AlignmentFlag.AlignBottom | Qt.AlignmentFlag.AlignRight,
            font=MuphynFonts.IOConnectorFont, parent=self)

        # Hide label
        self._lbl.hide()

    # -------------
    # Properties
    # -------------       
    @property
    def is_connected_to_input (self) -> bool :
        """Permet de savoir si l'élément actuel est connecté à une entrée (ou est un entrée)."""
        return True

    @property 
    def is_input (self) -> bool :
        """Permet de savoir si l'élément est une entrée."""
        return True

    # -------------
    # Methods
    # -------------
    def connector_bounding_rect(self):
        return self.connector_group.boundingRect()

    # -------------
    # Event Methods
    # -------------
    def connectorHoverEnterEvent(self, event: QGraphicsSceneHoverEvent) -> None:
        self._lbl.show()
        return super().hoverEnterEvent(event)

    def connectorHoverLeaveEvent(self, event: QGraphicsSceneHoverEvent) -> None:
        self._lbl.hide()
        return super().hoverLeaveEvent(event)
        
    def dragEnterEvent (self, event: QGraphicsSceneDragDropEvent) -> None :
        if event.possibleActions() == Qt.LinkAction :
            if not(self.is_connected) :
                if self.data_type.__str__() == event.mimeData().data('type') :
                    event.accept()
                    return super().dragEnterEvent(event)
            
        event.ignore()

        return super().dragEnterEvent(event)

    def dragMoveEvent(self, event : QGraphicsSceneDragDropEvent) -> None :
        if event.possibleActions() == Qt.LinkAction :
            if not(self.is_connected) :
                if self.data_type.__str__() == event.mimeData().data('type') :
                    event.accept()
                    return super().dragEnterEvent(event)
        
        event.ignore()

        return super().dragEnterEvent(event)

    def dropEvent (self, event: QGraphicsSceneDragDropEvent) -> None :
        if event.possibleActions() == Qt.LinkAction and not(self.is_connected) :
            if event.mimeData().data('action') == 'new link' :
                node = ctypes.cast(int(event.mimeData().data('link')), ctypes.py_object).value
                action = DiagramLinkNodesAction(self, node)
                action.do()
                self.scene().parent().parent().actions_holder.append(action)
        
                event.accept()

        return super().dropEvent(event)

    def add_link (self, link : Any) -> None :
        """Permet d'ajouter un lien à la position données."""

        if link is None : 
            return

        if len(self._links) > 0 :
            return

        self._links.append(link)

    def insert_link (self, index : int, link : Any) -> None :
        """Permet d'insérer un lien à la position données."""

        if link is None : 
            return

        if len(self._links) > 0 :
            return

        if index > len(self._links) :
            return

        self._links.insert(index, link)

    def has_link(self) -> bool:
        return len(self._links) > 0

    def to_dict(self) -> dict:
        input_dict = {
            "text": self.text,
            "signal_index": -1
        }

        return input_dict

class InputConnectionGroupModel(AbstractGraphicalElement):
    
    def __init__(self, name: str, is_infinite: bool, data_type: DataType, minimum_count: int, maximum_count: int, 
            default_count: int=0, group_position: QPointF = QPointF(), rotation: float = 0, text: str = '', parent: QGraphicsItem = None) -> None:
        super().__init__(name, group_position, rotation, text, parent)

        # Save group parameters
        self._name: str = name
        self._is_infinite = is_infinite
        self._data_type: DataType = data_type
        self._minimum_count: int = minimum_count if minimum_count >= 0 else 0
        self._maximum_count: int = maximum_count if maximum_count >= minimum_count else minimum_count+1

        # Init inputs list
        self._inputs: list[InputConnectionModel] = []

        for input_index in range(default_count):
            new_input = InputConnectionModel(
                name, 
                data_type, 
                QPointF(0, input_index * (AbstractConnectionModel.ItemHeight + 10)), 
                f"{name}_{input_index}", 
                is_infinite, 
                parent=self
            )
            self._inputs.append(new_input)

    
    # -------------
    # Properties
    # -------------
    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, new_name: str) -> None:
        if new_name != self._name:
            self._name = new_name

    @property
    def is_infinite(self) -> bool:
        return self._is_infinite

    @is_infinite.setter
    def is_infinite(self, new_is_infinite: str) -> None:
        if new_is_infinite != self._is_infinite:
            self._is_infinite = new_is_infinite

    @property
    def data_type(self) -> DataType:
        return self._data_type

    @data_type.setter
    def data_type(self, new_data_type: str) -> None:
        if new_data_type != self._data_type:
            self._data_type = new_data_type

    @property
    def minimum_count(self) -> int:
        return self._minimum_count

    @minimum_count.setter
    def minimum_count(self, new_minimum_count: str) -> None:
        if new_minimum_count != self._minimum_count:
            self._minimum_count = new_minimum_count

    @property
    def maximum_count(self) -> int:
        return self._maximum_count

    @maximum_count.setter
    def maximum_count(self, new_maximum_count: str) -> None:
        if new_maximum_count != self._maximum_count:
            self._maximum_count = new_maximum_count

    @property
    def count(self) -> int:
        return len(self._inputs)

    @count.setter
    def count(self, new_count: str) -> None:
        if new_count != self._count:
            self._count = new_count

    @property
    def inputs(self) -> list[InputConnectionModel]:
        return self._inputs

    @property
    def diagram_model (self) :
        return self.parent().diagram_model

    @diagram_model.setter
    def diagram_model (self, new_diagram_model) :
        self.diagram_model = new_diagram_model

    # -------------
    # Methods
    # -------------
    def connectors_bounding_rect(self):
        if len(self._inputs) == 0:
            return QRectF()
        elif len(self._inputs) == 1:
            return self._inputs[0].connector_bounding_rect()
        else:
            # Init connectors bounding rect
            connectors_bounding_rect = self._inputs[0].connector_bounding_rect()
            for input_ in self._inputs:
                connectors_bounding_rect = connectors_bounding_rect.united(input_.connector_bounding_rect())

            return connectors_bounding_rect

    def append_input(self) -> InputConnectionModel:
        # Add new graphical item in the group
        new_input = InputConnectionModel(
            self._name, 
            self._data_type, 
            QPointF(0, len(self._inputs) * (AbstractConnectionModel.ItemHeight + 10)), 
            f"{self._name}_{len(self._inputs)}", 
            self._is_infinite, 
            parent=self
        )
        new_input.action_param_semaphore = False

        # Append input in the list
        self._inputs.append(new_input)

        return new_input

    def insert_input(self, input_index) -> InputConnectionModel:
        # Recompute all inputs positions
        new_input = InputConnectionModel(
            self._name,
            self._data_type,
            QPointF(0, len(self._inputs) * (AbstractConnectionModel.ItemHeight + 10)),
            f"{self._name}_{len(self._inputs)}",
            self._is_infinite,
            parent=self
        )
        new_input.action_param_semaphore = False
        
        # Recompute all inputs positions
        self._inputs.insert(input_index, new_input)

        # Recompute all inputs positions
        self.recompute(from_index=input_index)

        return new_input

    @overload
    def remove_input(self, input_index: int) -> InputConnectionModel:
        ...

    @overload
    def remove_input(self, input_: InputConnectionModel) -> InputConnectionModel:
        ...

    def remove_input(self, input_: Union[int, InputConnectionModel]) -> InputConnectionModel:
        if type(input_) == int:
            # Get input_ index
            input_index = input_
            
            # Get input_ item
            input_ = self._inputs[input_index]
        elif type(input_) == InputConnectionModel:
            # Get input_ index
            input_index = self._inputs.index(input_)
            
            # Get input_ item
            input_ = self._inputs[input_index]
        else:
            LogManager().error(f"InputConnectionGroupModel.remove_input(): given arg is not a supported type '{type(input_)}' instead of 'int' or 'InputConnectionModel'")
            return 
            
        # Remove item from list
        self._inputs.remove(input_)

        # Remove from connection group
        input_.setParentItem(None)

        # Delete graphical item
        input_.deleteLater()

        # Recompute positions
        self.recompute(from_index=input_index)

    def clear_inputs(self):
        if self._is_infinite:
            # Remove only if this group is infinite
            while len(self._inputs) > 0:
                # Pop input_ from the list
                input_ = self._inputs.pop(0)

                # Remove graphical item
                input_.deleteLater()


    def recompute(self, from_index: int = 0):
        # If index is out of list limit → don't do anything
        if from_index < 0 or from_index >= len(self._inputs):
            return

        # Recompute positions
        for input_index, input_ in enumerate(self._inputs[from_index:]):
            input_.setPos(QPointF(0, input_index * (AbstractConnectionModel.ItemHeight + 10)))

    def to_dict(self) -> dict:
        inputs_group_dict = {
            "name": self.name,
            "isInfinite": self.is_infinite,
            "data_type": str(self.data_type),
            "inputs": [input_.to_dict() for input_ in self._inputs]
        }

        return inputs_group_dict


#-----------------------------------
# Imports
#-----------------------------------

from PyQt5.QtCore import QCoreApplication
from PyQt5.QtWidgets import QWidget, QLabel, QHBoxLayout
from muphyn.packages.interface.properties_pages.abstract_properties_editor import AbstractPropertiesEditor
from muphyn.packages.interface.widgets.BaseWidgets.Label import TitlePropertyLabel

#-----------------------------------
# Imports
#-----------------------------------

class TitlePropertiesElement (AbstractPropertiesEditor) :
    """Est la classe qui permet d'afficher un titre dans la liste des propriétées."""

    # -------------
    # Constructors
    # -------------
    
    def __init__ (self, title : str, parent : QWidget = None) :

        self._title = title

        AbstractPropertiesEditor.__init__(self, None)
        
    # -------------
    # Properties
    # -------------

    @property
    def title (self) -> str :
        """Permet de récuperer le titre affiché."""
        return self._title 

    # -------------
    # Methods
    # -------------

    def init_ui (self) -> None : 

        self._lbl_title: TitlePropertyLabel = TitlePropertyLabel()
        self.layout().addWidget(self._lbl_title)

    def translate_ui (self) -> None :
        
        self._lbl_title.setText(QCoreApplication.translate(self.objectName(), self.title, None))

    def create_layout(self) -> None:
        return QHBoxLayout()


from typing import Any
from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtWidgets import QWidget

class AbstractPropertyWidget(QWidget):

    valueChanged = pyqtSignal()

    def __init__(self, parent = None, flags = Qt.Widget) -> None:
        super().__init__(parent, flags)

        self._value = None

    
    # -------------
    # Properties
    # -------------
    @property
    def value(self) -> Any:
        return self._value

    @value.setter
    def value(self, new_value: Any):
        # Set new value
        self.setValue(new_value)


    # -------------
    # Methods
    # -------------
    def setValue(self, new_value: Any):
        raise(NotImplementedError(f"{type(self).__name__}.setValue() not yet implemented"))


import sys

from PyQt5 import QtWidgets, QtCore
from PyQt5.QtCore import Qt, QSize, QPoint
from PyQt5 import QtWidgets
from PyQt5.QtGui import *
from PyQt5 import QtGui

from PyQt5.QtWidgets import *


class Diedrico (QtWidgets.QWidget) :
    
    def paintEvent(self, event):
        qp = QtGui.QPainter(self)
        pen = QtGui.QPen(QtGui.QColor(QtCore.Qt.black), 5)
        qp.setPen(pen)
        qp.drawRect(500, 500, 1000, 1000)
        qp.drawRect(100, 100, 10, 20)
        qp.drawText(100, 100, "Test")

class ScrollArea(QWidget):

    _style = '''
            QScrollArea{
                background: white;
            }
            
            QScrollBar:handle{
                background: gray;
                max-width: 20px;
                color:green;
            
            }
            '''

    factor = 1.5

    def __init__(self, parent=None):
        super(ScrollArea, self).__init__()

        self.v_layout = QVBoxLayout(self)
        self.v_layout.setContentsMargins(0, 0, 0, 0)
        self.v_layout.setSpacing(0)

        self.container_widget = Diedrico()

        # Scroll Area Properties
        self.scroll = QScrollArea()
        self.scroll.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.scroll.setWidgetResizable(False)
        self.scroll.setWidget(self.container_widget)
        self.scroll.setStyleSheet(ScrollArea._style)
      
        self.container_widget.setGeometry(0, 0, self.width(), self.height())


        self.v_layout.addWidget(self.scroll)
        self.setLayout(self.v_layout)

        self._zoom = 0
        self.mousepos = QPoint(0, 0)
        self.setMouseTracking(True)
        self.showMaximized()

    def fitInView(self, scale=True):
        rect = QtCore.QRectF(self._photo.pixmap().rect())
        if not rect.isNull():
            self.setSceneRect(rect)
            if self.hasPhoto():
                unity = self.transform().mapRect(QtCore.QRectF(0, 0, 1, 1))
                self.scale(1 / unity.width(), 1 / unity.height())
                viewrect = self.viewport().rect()
                scenerect = self.transform().mapRect(rect)
                factor = min(viewrect.width() / scenerect.width(),
                             viewrect.height() / scenerect.height())
                self.scale(factor, factor)
            self._zoom = 0

    def wheelEvent(self, wheel_event):

        if wheel_event.modifiers() == Qt.ControlModifier:
            delta = wheel_event.angleDelta().y()
            if delta > 0:
                self.zoom_in()

            elif delta < 0:
                self.zoom_out()

        else:
            return super().wheelEvent(wheel_event)

    def mousePressEvent(self, event):
        cursor = self.container_widget.cursor().pos()
        print(cursor)
        if event.button() == Qt.MidButton:
            self.setCursor(Qt.OpenHandCursor)

        super(ScrollArea, self).mousePressEvent(event)

    def mouseMoveEvent(self, event):

        delta = event.localPos() - self.mousepos

        # panning area
        if event.buttons() == Qt.MidButton:
            h = self.scroll.horizontalScrollBar().value()
            v = self.scroll.verticalScrollBar().value()

            self.scroll.horizontalScrollBar().setValue(int(h - delta.x()))
            self.scroll.verticalScrollBar().setValue(int(v - delta.y()))

        self.mousepos = event.localPos()

        super(ScrollArea, self).mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):

        self.unsetCursor()
        self.mousepos = event.localPos()
        super(ScrollArea, self).mouseReleaseEvent(event)

    def resizeEvent(self, event):
        self.container_widget.resize(self.width(), self.height())

        super(ScrollArea, self).resizeEvent(event)

    def resize_container(self, option):

        option = int(option)

        if option == 0:
            self.container_widget.resize(self.width()+50, self.height())

        elif option == 1:
            self.container_widget.resize(self.width()+50, self.height())

        elif option == 2:
            self.container_widget.resize(self.width()+50, self.height()+50)

    @QtCore.pyqtSlot()
    def zoom_in(self):
        self.container_widget.setGeometry(200, 200, self.container_widget.width() + 4,
                                          self.container_widget.height() + 4)

    @QtCore.pyqtSlot()
    def zoom_out(self):
        self.container_widget.setGeometry(0, 0, self.container_widget.width() - 4,
                                          self.container_widget.height() - 4)


if __name__ == '__main__':
    a = QtWidgets.QApplication(sys.argv)
    q = ScrollArea()
    q.show()
    sys.exit(a.exec_())


#-----------------------------------
# Imports
#-----------------------------------


import os
from typing import List, Any

from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import pyqtSlot
from PyQt5.QtWidgets import QAbstractItemView, QFileDialog, QListWidgetItem, QVBoxLayout, QGridLayout

from muphyn.packages.core.box_library.plci_core_boxes_libraries import BoxesLibrariesManager
from muphyn.packages.core.scheduler_library.plci_core_schedulers_libraries import SchedulersLibraries
from muphyn.packages.core.utils.log_manager import LogManager
from muphyn.packages.interface.graphical_actions.actions_holder import ActionsHolder
from muphyn.packages.interface.dialogs.abstract_dialog import AbstractDialog
from muphyn.packages.interface.graphical_actions.library_dialog_add_action import LibraryDialogAddAction, _lst_contains_element
from muphyn.packages.interface.graphical_actions.library_dialog_remove_action import LibraryDialogRemoveAction
from muphyn.packages.interface.widgets.BaseWidgets.Button import PlainButton
from muphyn.packages.interface.utils.interface_constants import ApplicationWindowTitle
from muphyn.packages.interface.widgets.BaseWidgets.Label import TitlePropertyLabel

from muphyn.utils.paths import ROOT_DIR

#-----------------------------------
# Class
#-----------------------------------

class LibraryDialog (AbstractDialog) :
    """Est la classe permettant d'afficher une boîte de dialogue capable de modifier les bibliothèques des boxes et des solveurs."""
    
    # -------------
    # Constructors
    # -------------

    def __init__ (self, dialog_holder : Any, schedulers_libraries = SchedulersLibraries) :
        AbstractDialog.__init__(self, dialog_holder, 'library', f'{ApplicationWindowTitle} - Libraries')
        
        self._actions_holder = ActionsHolder()
        self.setMinimumSize(480, 360)
        self.resize(640, 480)
        self._init_ui()
        self._lst_libraries_selection_changed_event()
        self._fld_folder_text_edited_event()

        self._boxes_at_start : List[str] = []
        self._schedulers_libraries : SchedulersLibraries = schedulers_libraries
        self._solvers_at_start : List[str] = []

        for library_element in BoxesLibrariesManager().libraries :
            self._boxes_at_start.append(library_element.path)
            if not _lst_contains_element(self._lst_libraries, library_element.path) :
                QListWidgetItem(library_element.path, self._lst_libraries)

        for library_element in self._schedulers_libraries.libraries :
            self._solvers_at_start.append(library_element.path)
            if not _lst_contains_element(self._lst_libraries, library_element.path) :
                QListWidgetItem(library_element.path, self._lst_libraries)

        
    # -------------
    # Methods
    # -------------

    def keyPressEvent(self, event: QtGui.QKeyEvent) -> None:
        super().keyPressEvent(event)

        if event.modifiers() == QtCore.Qt.ControlModifier:

            if event.key() == QtCore.Qt.Key_Z :
                self._actions_holder.undo()

            elif event.key() == QtCore.Qt.Key_Y :
                self._actions_holder.redo()

        elif event.modifiers() == QtCore.Qt.NoModifier :

            if event.key() == QtCore.Qt.Key_Delete :

                if self._lst_libraries.hasFocus():
                    self._remove_method()

    def _init_ui (self) -> None :
        """
        Est la méthode appelée pour déssiner l'interface de la boîte de dialogue
        capable d'afficher l'interface de la boîte de dialogue.
        """

        # Init main Layout
        mainLayout = QVBoxLayout()

        # Init Options Layout
        optionsLayout = QGridLayout()
        optionsLayout.setColumnMinimumWidth(1, 45)
        optionsLayout.setColumnMinimumWidth(2, 45)
        optionsLayout.setSpacing(5)

        # Init Buttons Layout
        buttonsLayout = QGridLayout()
        buttonsLayout.setColumnMinimumWidth(1, 90)
        buttonsLayout.setColumnMinimumWidth(2, 90)
        buttonsLayout.setHorizontalSpacing(5)

        # Folder Path Field Text
        self._fld_folder = QtWidgets.QLineEdit(self)
        self._fld_folder.setObjectName("_fld_folder")
        self._fld_folder.textEdited.connect(self._fld_folder_text_edited_event)

        # Search Button (Open a Folder Dialog)
        self._search_button = QtWidgets.QPushButton(self)
        self._search_button.setObjectName("_search_button")
        self._search_button.clicked.connect(self._search_method)

        # Library list title
        self._lbl_libraries = TitlePropertyLabel(self)
        self._lbl_libraries.setObjectName("_lbl_libraries")

        # Add Path Field Text Content to list as Item
        self._add_button = PlainButton(self)
        self._add_button.setObjectName("_add_button")
        self._add_button.clicked.connect(self._add_method)

        # Remove Item From List Button
        self._remove_button = QtWidgets.QPushButton(self)
        self._remove_button.setObjectName("_remove_button")
        self._remove_button.clicked.connect(self._remove_method)

        # List of all library pathes
        self._lst_libraries = QtWidgets.QListWidget()
        self._lst_libraries.setObjectName("_lst_libraries")
        self._lst_libraries.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self._lst_libraries.itemSelectionChanged.connect(self._lst_libraries_selection_changed_event)

        # Confirm Button (Update Boxes List)
        self._accept_button = PlainButton()
        self._accept_button.setObjectName("_accept_button")
        self._accept_button.clicked.connect(self._accept_method)

        # Cancel button (just close the library dialog and restore old library list)
        self._cancel_button = QtWidgets.QPushButton()
        self._cancel_button.setObjectName("_cancel_button")
        self._cancel_button.clicked.connect(self._cancel_method)

        # Adding options widget to layout
        optionsLayout.addWidget(self._fld_folder, 0, 0)
        optionsLayout.addWidget(self._search_button, 0, 1, 1, 2)
        optionsLayout.addWidget(self._lbl_libraries, 1, 0)
        optionsLayout.addWidget(self._add_button, 1, 1)
        optionsLayout.addWidget(self._remove_button, 1, 2)
        optionsLayout.addWidget(self._lst_libraries, 2, 0, 1, 3)
        optionsLayout.setRowStretch(2, 1)
        optionsLayout.setColumnStretch(0, 1)

        # Closing action layout
        buttonsLayout.addWidget(self._accept_button, 0, 1)
        buttonsLayout.addWidget(self._cancel_button, 0, 2)
        buttonsLayout.setColumnStretch(0, 1)

        # Main Layout
        mainLayout.addItem(optionsLayout)
        mainLayout.addItem(buttonsLayout)
        mainLayout.setSpacing(10)

        # Add Layout to file dialog
        self.setLayout(mainLayout)

        self._retranslate_ui(self)
        QtCore.QMetaObject.connectSlotsByName(self)

    def _retranslate_ui (self, Dialog) -> None :
        _translate = QtCore.QCoreApplication.translate
        
        self._search_button.setText(_translate("Dialog", "Select Folder"))
        self._add_button.setText(_translate("Dialog", "+"))
        self._remove_button.setText(_translate("Dialog", "-"))
        self._accept_button.setText(_translate("Dialog", "Apply"))
        self._cancel_button.setText(_translate("Dialog", "Cancel"))
        self._lbl_libraries.setText(_translate("Dialog", "Libraries"))

    @pyqtSlot()
    def _search_method (self) -> None :
        """Est la méthode appelée lorsque l'utilisateur veut faire une recherche."""
        path = QFileDialog.getExistingDirectory(self, f"{ApplicationWindowTitle} - Select Libraries Folder", ROOT_DIR)

        if path is None or path == '':
            return

        self._fld_folder.setText(path)
        self._fld_folder_text_edited_event()

    @pyqtSlot()
    def _add_method (self) -> None :
        """Est la méthode appelée pour ajouter un élément dans la liste."""

        if _lst_contains_element(self._lst_libraries, self._fld_folder.text()):
            return

        library_action = LibraryDialogAddAction(self._lst_libraries, self._fld_folder.text())
        library_action.do()
        self._actions_holder.append(library_action)

    @pyqtSlot()
    def _remove_method (self) -> None :
        """Est la méthode appelée pour supprimer un élément de la liste."""
        
        selected_items : List[str] = []

        for selected_item in self._lst_libraries.selectedItems() :
            selected_items.append(selected_item.text())

        library_action = LibraryDialogRemoveAction(self._lst_libraries, selected_items)
        library_action.do()
        self._actions_holder.append(library_action)

    @pyqtSlot()
    def _accept_method (self) -> None :
        """Est la méthode appelée lorsque l'utilisateur accepte les modifications apportées."""
        
        BoxesLibrariesManager().clear()
        self._schedulers_libraries.clear()

        for i in range(self._lst_libraries.__len__()) :
            path : str = self._lst_libraries.item(i).text()

            LogManager().debug(f"Adding \"{path}\"...")

            BoxesLibrariesManager().add_library(path)
            self._schedulers_libraries.add_library(path)

        BoxesLibrariesManager().load_libraries()
        self._schedulers_libraries.load_libraries()

        self.close()

    @pyqtSlot()
    def _cancel_method (self) -> None :
        """Est la méthode appelée lorsque l'utilisateur annulle les modifications apportées."""
        self.close()

    @pyqtSlot()
    def _lst_libraries_selection_changed_event (self) -> None :
        """Est la méthode appelée lorsque la sélection des bibliothèque change."""
        self._remove_button.setEnabled(self._lst_libraries.selectedItems().__len__() > 0)
    
    @pyqtSlot()
    def _fld_folder_text_edited_event (self) -> None :
        """Est la méthode appelée lorsque le texte dans le champ de sélection de dossier est modifié."""
        if self._fld_folder.text().__len__() > 0:
            self._add_button.setEnabled(os.path.isdir(self._fld_folder.text()))

        else:
            self._add_button.setEnabled(False)


#-----------------------------------
# Imports
#-----------------------------------
import re
from typing import List, Dict

from muphyn.packages.core.box_library.plci_core_boxes_libraries import BoxesLibrariesManager
from muphyn.packages.core.plci_core_box import Box
from muphyn.packages.core.plci_core_diagram import Diagram
from muphyn.packages.core.plci_core_scheduler import Scheduler
from muphyn.packages.core.plci_core_scheduler_params import SchedulerParams
from muphyn.packages.core.plci_core_signal import Signal
from muphyn.packages.core.scheduler_library.plci_core_schedulers_libraries import SchedulersLibraries
from muphyn.packages.interface.models.editable_models.scheduler_model import SchedulerModel
from muphyn.packages.interface.models.editable_models.simulation_model import SimulationModel
from muphyn.packages.interface.models.graphical_models.box_model import BoxModel
from muphyn.packages.interface.models.signals_model.abstract_signal_model import AbstractSignalModel
from muphyn.packages.core.utils.log_manager import LogManager
from muphyn.packages.core.utils.global_env_variables_manager import GlobalEnvVariablesManager
from muphyn.packages.core.utils.regexes import *
from muphyn.packages.core.plci_core_data_type import get_data_type

#-----------------------------------
# Class
#-----------------------------------

class Parser : 

    # -------------
    # Constructors
    # -------------

    def __init__ (self) :
        ...
    
    # -------------
    # Methods
    # -------------      

    def parse (self, element : object, schedulers_libraries : SchedulersLibraries) -> Scheduler :
        """Est la méthode pour convertir/traduire un objet en un scheduler prêt à être simuler."""
        
        if isinstance(element, SimulationModel) :
            
            return self.__parse_simulation__(element, BoxesLibrariesManager(), schedulers_libraries)

        elif isinstance(element, str) :
            ...

        return None

    def __parse_simulation__ (self, simulation_model : SimulationModel, boxes_libraries : BoxesLibrariesManager, schedulers_libraries : SchedulersLibraries) -> Scheduler :
        """Est la méthode pour convertir/traduire une simulation en un scheduler prêt à être simuler."""

        if simulation_model.scheduler_model is None :
            return None

        else :
            if simulation_model.scheduler_model.params is None :
                return None 

            else :

                scheduler = schedulers_libraries.construct_scheduler(simulation_model.scheduler_model.library, simulation_model.scheduler_model.name)

                # Si pas de scheduler sélectionné.
                if scheduler is None : 
                    
                    scheduler_library = None 

                    # On prend le premier.
                    for lib in schedulers_libraries.schedulers :
                        if not(lib is None) :
                            scheduler_library = lib
                            break

                    if scheduler_library is None :
                        return None

                    else :
                        simulation_model.scheduler_model = SchedulerModel(scheduler_library.scheduler_library, scheduler_library.scheduler_name, simulation_model.scheduler_model.params)
                        scheduler = scheduler_library.construct_scheduler()

                scheduler.params = SchedulerParams(simulation_model.scheduler_model.params.stop_time, simulation_model.scheduler_model.params.step_time)
                scheduler.diagram = Diagram()
                
                last_signal_index = 0
                signals_dict : Dict[AbstractSignalModel, Signal] = {} 
                diagrams : List[Diagram] = []

                for box_model in simulation_model.box :

                    if isinstance(box_model, BoxModel) :

                        params = {}
                        for param in box_model.get_parameters() :
                            params[param] = box_model.get_parameter(param)['value']

                        box = boxes_libraries.construct_box(box_model.library, box_model.name, **params)
                        scheduler.diagram.append(box)

                        if isinstance(box, Box) :

                            # Handle Inputs
                            for input in box_model.inputs :

                                if len(input._links) > 0 :
                                    link = input._links[0]

                                    if link in signals_dict :
                                        scheduler.diagram.add_box_inputs(box, signals_dict[link])
                                        signals_dict[link].input_name = input.text

                                    else :
                                        signal : Signal = Signal(last_signal_index, link.data_type, link.data_type.default_value(), input_name=input.text)
                                        scheduler.diagram.append(signal)
                                        signals_dict[link] = signal
                                        scheduler.diagram.add_box_inputs(box, signal)
                                        last_signal_index += 1
                                        
                            # Handle Outputs
                            for output in box_model.outputs :
                                
                                for link in output.links : 
                                    
                                    if link in signals_dict :
                                        scheduler.diagram.add_box_outputs(box, signals_dict[link])
                                        signals_dict[link].ouput_name = output.text

                                    else :
                                        signal : Signal = Signal(last_signal_index, link.data_type, link.data_type.default_value(), output_name=output.text)
                                        scheduler.diagram.append(signal)
                                        signals_dict[link] = signal
                                        scheduler.diagram.add_box_outputs(box, signal)
                                        last_signal_index += 1

                            # Handle Parameters
                            for param in box.params:
                                # Get param value from Box
                                param_value = box.get_parameter(param)
                                if type(param_value) == str:
                                
                                    # Test if scientific notation number
                                    if re.match(DotScientificNumberRegex, param_value):
                                        # Valid Value
                                        box[param] = float(param_value)

                                    # Test if value exists as global variable
                                    elif str(param_value) in GlobalEnvVariablesManager().global_vars:
                                        # Get global variable value
                                        global_var = GlobalEnvVariablesManager().global_vars[param_value]

                                        # Get parameter type
                                        param_type = get_data_type(box_model.get_parameter(param)["type"])
                                        
                                        # Compare types
                                        if str(param_type) == "float" and str(type(global_var).__name__) in ['float', 'int']:
                                            box[param] = float(global_var)
                                        elif str(param_type) == "int" and str(type(global_var).__name__) in ['float', 'int']:
                                            box[param] = int(global_var)
                                        else:
                                            box[param] = param_type.default_value()
                                            

                        elif isinstance(box, Diagram) : 
                            diagrams.append(box)
                            

                return scheduler

                        


from copy import deepcopy
from enum import Enum
from typing import Iterable, overload, Union

from PyQt5.QtCore import QLine, QLineF, QPointF, QRectF, QSize,  QSizeF, Qt, pyqtSignal
from PyQt5.QtGui import QColor, QGradient, QFont, QFontMetricsF, QGradient, QPainter, QPainterPath
from PyQt5.QtWidgets import QStyleOptionGraphicsItem, QApplication

from muphyn.packages.core.utils.log_manager import LogManager

from muphyn.packages.interface.shapes.basic_shape.abstract_shape import AbstractShape

class Circle(AbstractShape):

    center_changed = pyqtSignal()
    radius_changed = pyqtSignal()

    def __init__(self, center: QPointF, radius: float, border_color: QColor = Qt.GlobalColor.black, border_width: int = 1, 
                    line_style: Qt.PenStyle = Qt.PenStyle.SolidLine, fill_color: Union[QColor, QGradient] = Qt.GlobalColor.transparent, parent=None):

        # Calc Pos & Size
        origin = Circle.calc_pos(center, radius)
        size = Circle.calc_size(center, radius)

        # Center (reset center position from origin)
        self._center = center - origin

        # Radius
        self._radius = radius


        super().__init__(origin, size, border_color, border_width, line_style, fill_color, parent=parent)

    # -------------
    # Properties
    # -------------
    @property
    def center(self) -> QPointF:
        return self._center

    @center.setter
    def center(self, new_center: QPointF) -> None:
        # Calc new origin
        new_origin = Circle.calc_pos(new_center, self._radius)

        if new_origin != self.pos():
            # Change position
            self.setPos(new_origin)
            self.position_changed.emit()

    @property
    def radius(self) -> float:
        return self._radius

    @radius.setter
    def radius(self, new_radius: float) -> None:
        if new_radius != self._radius:
            # Set new center
            self._center = QPointF(2*new_radius, 2*new_radius)

            # Set new radius
            self._radius = new_radius

            # Calc new origin & size
            new_origin = Circle.calc_pos(self._center, new_radius)
            new_size = Circle.calc_size(self._center, new_radius)

            # Set pos
            if new_origin != self.pos():
                # Change position
                self.setPos(new_origin)
                self.position_changed.emit()

            # Set size
            if new_size != self._size:
                self._size = new_size
                self.size_changed.emit()

    # -------------
    # Methods
    # -------------
    def draw_shape(self, painter: QPainter, option: QStyleOptionGraphicsItem, widget) -> None:
        painter.drawEllipse(self._center, self._radius, self._radius)
        
    # --------------
    # Static Methods
    # --------------
    @staticmethod
    def calc_pos(center: QPointF, radius: float) -> QPointF:
        top_left = center - QPointF(radius, radius)
        return top_left

    @staticmethod
    def calc_size(center: QPointF, radius: float) -> QSizeF:
        size = QSizeF(2*radius, 2*radius)
        return size

class Ellipse(AbstractShape):

    center_changed = pyqtSignal()
    rx_changed = pyqtSignal()
    ry_changed = pyqtSignal()

    def __init__(self, center: QPointF, rx: float, ry: float, border_color: QColor, border_width: int = 1, 
                    line_style: Qt.PenStyle = Qt.PenStyle.SolidLine, fill_color: Union[QColor, QGradient] = Qt.GlobalColor.transparent, parent=None):

        # Calc Pos & Size
        origin = Ellipse.calc_pos(center, rx, ry)
        size = Ellipse.calc_size(center, rx, ry)

        # Center (reset center position from origin)
        self._center = center - origin

        # Radius
        self._rx = rx
        self._ry = ry

        super().__init__(origin, size, border_color, border_width, line_style, fill_color, parent=parent)

    # -------------
    # Properties
    # -------------
    @property
    def center(self) -> QPointF:
        return self._center

    @center.setter
    def center(self, new_center: QPointF) -> None:
        # Calc new origin
        new_origin = Ellipse.calc_pos(new_center, self._radius)

        if new_origin != self.pos():
            # Change position
            self.setPos(new_origin)
            self.position_changed.emit()

    @property
    def rx(self) -> float:
        return self._rx

    @rx.setter
    def rx(self, new_rx: float) -> None:
        if new_rx != self._rx:
            # Set new center
            self._center = QPointF(2*new_rx, 2*self._ry)

            # Set new radius
            self._rx = new_rx

            # Calc new origin & size
            new_origin = Ellipse.calc_pos(self._center, new_rx, self._ry)
            new_size = Ellipse.calc_size(self._center, new_rx, self._ry)

            # Set pos
            if new_origin != self.pos():
                # Change position
                self.setPos(new_origin)
                self.position_changed.emit()

            # Set size
            if new_size != self._size:
                self._size = new_size
                self.size_changed.emit()

    # -------------
    # Methods
    # -------------
    def draw_shape(self, painter: QPainter, option: QStyleOptionGraphicsItem, widget) -> None:
        painter.drawEllipse(self._center, self._rx, self._ry)
        
    # --------------
    # Static Methods
    # --------------
    @staticmethod
    def calc_pos(center: QPointF, rx: float, ry: float) -> QPointF:
        top_left = center - QPointF(rx, ry)
        return top_left

    @staticmethod
    def calc_size(rx: float, ry: float) -> QSizeF:
        size = QSizeF(2*rx, 2*ry)
        return size

class Line(AbstractShape):
    def __init__(self, line: Union[QLine, QLineF], 
                    line_color: QColor = Qt.GlobalColor.black, line_width: int = 1, 
                    line_style: Qt.PenStyle = Qt.PenStyle.SolidLine, parent=None):
        
        # Line format handle
        if not isinstance(line, QLineF):
            line = QLineF(line)

        # Calc Pos & Size
        origin = Line.calc_pos(line)
        size = Line.calc_size(line)

        # Reset line from object origin
        self._line = line
        self._from_origin_line = line.translated(-origin)

        super().__init__(origin, size, line_color, line_width, line_style, parent=parent)

    # -------------
    # Properties
    # -------------
    @property
    def line(self) -> QPointF:
        return self._from_origin_line

    @line.setter
    def line(self, new_line: Union[QLine, QLineF]) -> None:
        if not isinstance(new_line, QLineF):
            new_line = QLineF(new_line)

        if self._line != new_line:
            # Set new line
            self._line = new_line

            # Calc from origin line
            self._from_origin_line = new_line.translated(-new_origin)

            # Calc new origin & size
            new_origin = Line.calc_pos(new_line)
            new_size = Line.calc_size(new_line)

            # Set Pos
            if new_origin != self.pos():
                self.setPos(new_origin)
                self.position_changed.emit()

            # Set size
            self.setOrigin(new_origin)
            if new_size != self._size:
                self._size = new_size
                self.size_changed.emit()

    # -------------
    # Methods
    # -------------
    def draw_shape(self, painter: QPainter, option: QStyleOptionGraphicsItem, widget) -> None:
        painter.drawLine(self._from_origin_line)
    
    # --------------
    # Static Methods
    # --------------
    @staticmethod
    def calc_pos(line: Union[QLine, QLineF]) -> QPointF:
        # Calculate top left
        top_left_x = min(line.x1(), line.x2())
        top_left_y = min(line.y1(), line.y2())

        return QPointF(top_left_x, top_left_y)

    @staticmethod
    def calc_size(line: Union[QLine, QLineF]) -> QSizeF:
        # Calculate top left
        top_left_x = min(line.x1(), line.x2())
        top_left_y = min(line.y1(), line.y2())

        # Calculate bottom right
        bottom_right_x = max(line.x1(), line.x2())
        bottom_right_y = max(line.y1(), line.y2())

        width = abs(bottom_right_x - top_left_x)
        height = abs(bottom_right_y - top_left_y)

        return QSizeF(width, height)

class Path(AbstractShape):

    class StepType(Enum):
        Line = 1
        Cubic = 2

    class Step():

        @overload
        def __init__(self, point: QPointF) -> None:
            ...

        @overload
        def __init__(self, c1: QPointF, c2: QPointF, end_point: QPointF) -> None:
            ...

        def __init__(self, *args: Iterable[QPointF]) -> None:
            if len(args) == 1 and type(args[0]) == QPointF:
                # Step Type
                self.type = Path.StepType.Line

                self.point = args[0]

            elif len(args) == 3 and type(args[0]) == QPointF and type(args[1]) == QPointF and type(args[2]) == QPointF:
                # Step Type
                self.type = Path.StepType.Cubic

                # 
                self.c1 = args[0]
                self.c2 = args[1]
                self.end_point = args[2]

            else:
                # StepType.Line Step object building error
                if len(args) != 1:
                    missing_args = 1 - len(args)
                    line_object_error = f"Wrong number of args : {abs(missing_args)} {'missing_values' if missing_args > 0 else 'in excess'}"
                else:
                    line_object_error = f"Wrong type given: {type(args[0])} instead of QPointF"

                if len(args) != 3:
                    missing_args = 3 - len(args)
                    cubic_object_error = f"Wrong number of args : {abs(missing_args)} {'missing_values' if missing_args > 0 else 'in excess'}"
                else:
                    type_errors = []
                    if type(args[0]) != QPointF:
                        type_errors.append(f"c1 : {type(args[0])} instead of QPointF")
                    if type(args[1]) != QPointF:
                        type_errors.append(f"c1 : {type(args[1])} instead of QPointF")
                    if type(args[2]) != QPointF:
                        type_errors.append(f"c1 : {type(args[2])} instead of QPointF")
                    
                    cubic_object_error = f"Wrong type given: ({', '.join(type_errors)})"

                constructor_args_errors = '\n'.join([
                    f" - Step(point: QPointF) -> {line_object_error}",
                    f" - Step(c1: QPointF, c2: QPointF, end_point: QPointF) -> {cubic_object_error}"
                ])

                raise(NotImplementedError(
                        f"Wrong args given :\n{constructor_args_errors}"
                    )
                )


    def __init__(self, steps: Iterable[Step], line_color: QColor = Qt.GlobalColor.black, 
                    line_width: int = 1, line_style: Qt.PenStyle = Qt.PenStyle.SolidLine, 
                    fill_color: Union[QColor, QGradient] = Qt.GlobalColor.transparent, pen_join_style: Qt.PenJoinStyle = Qt.PenJoinStyle.MiterJoin, parent=None):

        # Handle points
        if len(steps) < 2:
            LogManager().error(f"Path.__init__: not enough steps given to create a polygon, must give at least 2 steps instead of {len(steps)}")
            return

        self._steps = steps

        # Join Style
        self._pen_join_style = pen_join_style

        # Painter path
        self._painter_path = Path.build_path(steps)

        # Calc origin and size
        bounding_rect = self._painter_path.boundingRect()
        origin = bounding_rect.topLeft()
        size = bounding_rect.size()

        # Save old bounding rectangle
        self._old_bounding_rect = self._painter_path.boundingRect()
        
        super().__init__(origin, size, line_color, line_width, line_style, fill_color, parent=parent)


    # -------------
    # Properties
    # -------------
    @property
    def steps(self) -> Iterable[Step]:
        return self._steps

    @steps.setter
    def steps(self, new_steps) -> None:
        # Save old bounding rectangle
        self._old_bounding_rect = self._painter_path.boundingRect()

        # Build new painter path
        self._painter_path = Path.build_path(new_steps)

        # Save new steps
        self._steps = new_steps

    # -------------
    # Methods
    # -------------
    def boundingRect(self) -> QRectF:
        return self._painter_path.boundingRect().united(self._old_bounding_rect)
        
    def draw_shape(self, painter: QPainter, option: QStyleOptionGraphicsItem, widget) -> None:
        painter.drawPath(self._painter_path)
        
    # --------------
    # Static Methods
    # --------------
    @staticmethod
    def build_path(steps: Iterable[Step]) -> QPainterPath:
        # Handling first step
        start_point = QPointF()
        first_step: Path.Step = steps[0]
        if first_step.type == Path.StepType.Line:
            start_point = first_step.point
        elif first_step.type == Path.StepType.Cubic:
            start_point = first_step.end_point
        else:
            raise(TypeError("Path.build_path(): Wrong step type given: {first_step.type} instead of Path.StepType.Line or Path.StepType.Cubic"))

        # Handling the rest of the path
        painter_path = QPainterPath(start_point)
        for step in steps:
            if step.type == Path.StepType.Line:
                painter_path.lineTo(step.point)
            elif step.type == Path.StepType.Cubic:
                painter_path.cubicTo(step.c1, step.c2, step.end_point)
            else:
                raise(TypeError("Path.build_path(): Wrong step type given: {step.type} instead of Path.StepType.Line or Path.StepType.Cubic"))

        return painter_path

class Polygon(AbstractShape):
    def __init__(self, points: Iterable[QPointF], border_color: QColor = Qt.GlobalColor.black, 
                    border_width: int = 1, line_style: Qt.PenStyle = Qt.PenStyle.SolidLine, 
                    fill_color: Union[QColor, QGradient] = Qt.GlobalColor.transparent, pen_join_style: Qt.PenJoinStyle = Qt.PenJoinStyle.MiterJoin, parent=None):

        # Handle points
        if len(points) < 3:
            LogManager().error(f"Polygon.__init__: not enough points given to create a polygon, must give at least 3 points instead of {len(points)}")
            return

        # Calc Pos & Size
        origin = self.calc_pos(points)
        size = self.calc_size(points)

        # Translate all points form the origin
        self._points = [point - origin for point in points]

        # Join Style
        self._pen_join_style = pen_join_style

        super().__init__(origin, size, border_color, border_width, line_style, fill_color, parent=parent)

    # -------------
    # Properties
    # -------------
    @property
    def points(self) -> Iterable[QPointF]:
        return self._points

    @points.setter
    def points(self, new_points: Iterable[QPointF]) -> None:
        # Calc new origin & size
        new_origin = Polygon.calc_pos(new_points)
        new_size = Polygon.calc_size(new_points)

        # Set new points
        self._points = [point - new_origin for point in new_points]

        # Set Pos
        if new_origin != self.pos():
            self.setPos(new_origin)
            self.position_changed.emit()

        # Set size
        self.setOrigin(new_origin)
        if new_size != self._size:
            self._size = new_size
            self.size_changed.emit()

    @property
    def size(self) -> QSizeF:
        return self.boundingRect().size()

    
    # -------------
    # Methods
    # -------------
    def draw_shape(self, painter: QPainter, option: QStyleOptionGraphicsItem, widget) -> None:
        painter.drawPolygon(self._points)
        
    # --------------
    # Static Methods
    # --------------
    @staticmethod
    def calc_pos(points: Iterable[QPointF]) -> QPointF:
        # Calc bounding position
        min_x = min([point.x() for point in points])
        min_y = min([point.y() for point in points])

        return QPointF(min_x, min_y)

    @staticmethod
    def calc_size(points: Iterable[QPointF]) -> QSizeF:
        xs = [point.x() for point in points]
        ys = [point.y() for point in points]

        # Calc bounding position
        min_x = min(xs)
        min_y = min(ys)

        # Calc bounding position
        max_x = max(xs)
        max_y = max(ys)

        # Calc width & height
        width = abs(max_x - min_x)
        height = abs(max_y - min_y)

        return QSizeF(width, height)

class Rectangle(AbstractShape):

    def __init__(self, top_left: QPointF, width: float, height: float, border_color: QColor = Qt.GlobalColor.black, border_width: int = 1, 
                    line_style: Qt.PenStyle = Qt.PenStyle.SolidLine, fill_color: Union[QColor, QGradient] = Qt.GlobalColor.transparent, parent=None):
        
        # Calc Size
        size = QSizeF(width, height)

        super().__init__(top_left, size, border_color, border_width, line_style, fill_color, parent=parent)

    # -------------
    # Properties
    # -------------
    @property
    def width(self) -> float:
        return self._size.width()

    @width.setter
    def width(self, new_width: float):
        if new_width != self._size.width():
            self._size.setWidth(new_width)

    @property
    def height(self) -> float:
        return self._size.heigth()

    @height.setter
    def height(self, new_height: float):
        if new_height != self._size.heigth():
            self._size.setHeight(new_height)

    @property
    def size(self) -> QSizeF:
        return self._size

    @size.setter
    def size(self, new_size) -> None:
        if new_size != self._size:
            self._size = new_size
    
    # -------------
    # Methods
    # -------------
    def draw_shape(self, painter: QPainter, option: QStyleOptionGraphicsItem, widget) -> None:
        painter.drawRect(QRectF(QPointF(0, 0), self._size))

class Square(Rectangle):

    def __init__(self, top_left: QPointF, side_size: float, 
                    border_color: QColor, border_width: int = 1, line_style: Qt.PenStyle = Qt.PenStyle.SolidLine, 
                    fill_color: Union[QColor, QGradient] = Qt.GlobalColor.transparent, parent=None):
        
        super().__init__(top_left, side_size, side_size, border_color, 
                            border_width, line_style, 
                            fill_color, parent=parent)

    
    # -------------
    # Properties
    # -------------
    @property
    def side_size(self) -> float:
        return self._size.width()

    @side_size.setter
    def side_size(self, new_side_size: float):
        if new_side_size != self._size.width():
            self._size.setWidth(new_side_size)
            self._size.setHeight(new_side_size)

class Text(AbstractShape):

    def __init__(self, text: str, position: QPointF = QPointF(), 
                    text_max_width: float = 0.0, text_max_height: float = 0.0, 
                    font: QFont = None, font_color: QColor = Qt.GlobalColor.black, 
                    background_color: Union[QColor, QGradient] = QColor(0, 0, 0, 0), 
                    alignment: Qt.AlignmentFlag = Qt.AlignmentFlag.AlignTop | Qt.AlignmentFlag.AlignLeft, 
                    parent=None):

        # Text
        self._text = text

        # Text size parameters
        self._text_max_width = text_max_width
        self._text_max_height = text_max_height

        # Text option
        word_wrap = Qt.TextFlag.TextWordWrap if text_max_width > 0.0 else Qt.TextFlag.TextSingleLine
        self._text_flags = alignment | word_wrap

        # Font
        if font is None :
            font = QFont()
        self._font = font

        # Font metrics
        font_copy = QFont(font)
        font_copy.setPointSize(int(font.pointSize() * QApplication.desktop().logicalDpiX()/96.0))
        self._font_metrics = QFontMetricsF(font_copy)

        # Calculate rectangle
        bounding_rect = QRectF(position, QSizeF(text_max_width, text_max_height))
        bounding_rect = self._font_metrics.boundingRect(bounding_rect, self._text_flags, text)

        # Translate Rectangle
        if bounding_rect.y() < position.y():
            offset_y = position.y() - bounding_rect.y() 
            bounding_rect.translate(0, offset_y)

        # Reset drawing position from the origin
        self._drawing_rect = bounding_rect
        self._drawing_rect.translate(-position)

        # 
        super().__init__(position, bounding_rect.size(), font_color, background_color=background_color, parent=parent)

    # -------------
    # Properties
    # -------------
    @property
    def text(self) -> str:
        return self._text
    
    @text.setter
    def text(self, new_text) -> None:
        print(new_text,  self._text, new_text != self._text)
        if new_text != self._text:
            # Get old bounding rect
            old_bounding_rect = self.boundingRect()

            # Replace value
            self._text = new_text

            # Calculate text rectangle
            bounding_rect = QRectF(QPointF(0, 0), QSizeF(self._text_max_width, self._text_max_height))
            bounding_rect = self._font_metrics.boundingRect(bounding_rect, self._text_flags, new_text)
            
            # Set new bounding rect
            self._drawing_rect = bounding_rect

            self.update()

    # -------------
    # Methods
    # -------------
    def boundingRect(self) -> QRectF:
        return self._drawing_rect

    def draw_shape(self, painter: QPainter, option: QStyleOptionGraphicsItem, widget) -> None:
        # Set Font
        painter.setFont(self._font)

        # Painter set brush as transparent
        painter.setBrush(QColor(0, 0, 0, 0))

        # Draw Text
        painter.drawText(self._drawing_rect, self._text_flags, self._text)

    def setSize(self, new_size: Union[QSize, QSizeF]):
        # Handle new size type
        new_size = QSizeF(new_size) if type(new_size) == QSize else new_size

        if self._size != new_size:
            # Update size value
            self._size = new_size
            self._text_max_width = new_size.width()
            self._text_max_height = new_size.height()

            # Calculate text rectangle
            bounding_rect = QRectF(QPointF(0, 0), QSizeF(self._text_max_width, self._text_max_height))
            bounding_rect = self._font_metrics.boundingRect(bounding_rect, self._text_flags, self._text)
                
            self._drawing_rect = bounding_rect


#-----------------------------------
# Imports
#-----------------------------------

# General Imports
from typing import Dict

# PyQt5 Imports
from PyQt5.QtCore import QPointF, QSizeF
from PyQt5.QtGui import QColor
from muphyn.packages.core.box_library.plci_core_boxes_libraries import BoxesLibrariesManager

# Project Imports
from muphyn.packages.core.utils.log_manager import LogManager
from muphyn.packages.interface.models.graphical_models.abstract_box_model import AbstractBoxModel
from muphyn.packages.interface.models.graphical_models.abstract_graphical_element import AbstractGraphicalElement
from muphyn.packages.interface.models.graphical_models.box_input_model import BoxInputModel
from muphyn.packages.interface.models.graphical_models.box_model import BoxModel
from muphyn.packages.interface.models.graphical_models.box_output_model import BoxOutputModel
from muphyn.packages.interface.models.signals_model.signal_link_model import SignalLinkModel
from muphyn.packages.interface.models.editable_models.abstract_diagram_model import AbstractDiagramModel

#-----------------------------------
# Functions
#-----------------------------------

def color (color_dict : Dict) -> QColor :
    """Permet de reconstruire une couleur sur base d'un dictionnaire."""
    return QColor(color_dict['red'], color_dict['green'], color_dict['blue'])
    
def link (link_reconstructor : Dict, diagram_model : AbstractDiagramModel) -> SignalLinkModel :
    """Permet de reconstruire un lien suivant son reconstructeur."""
    input_box = diagram_model.get_element_by_graphical_index(link_reconstructor['input_box_index'])
    output_box = diagram_model.get_element_by_graphical_index(link_reconstructor['output_box_index'])

    if input_box is None : 
        LogManager().error('reconstruct link : input box = None')
        return None

    if not(hasattr(input_box, '_inputs')) :
        LogManager().error('reconstruct link : input box has no inputs')
        return None

    if output_box is None : 
        LogManager().error('reconstruct link : output box = None')
        return None

    if not(hasattr(output_box, '_outputs')) :
        LogManager().error('reconstruct link : output box has no inputs')
        return None

    input = None
    output = None

    for i in input_box.inputs : 
        if i.graphical_index == link_reconstructor['input_index'] :
            input = i
            break

    for o in output_box.outputs :
        if o.graphical_index == link_reconstructor['output_index'] : 
            output = o
            break

    if input is None :
        LogManager().error('reconstruct link : no input found')
        return None
        
    if output is None : 
        LogManager().error('reconstruct link : no output found')
        return None


    link_model = diagram_model.link_nodes(input, output, link_value = link_reconstructor['link_value'], link_type = link_reconstructor['link_type'], link_text = link_reconstructor['link_text']) 
    link_model.graphical_index = link_reconstructor['graphical_index']

    return link_model

def box (box_reconstructor : Dict) -> BoxModel :
    """Permet de reconstruire une box suivant son reconstructeur."""

    # Box data
    library = box_reconstructor['library']
    name = box_reconstructor['name']

    # Get box data
    box_data = BoxesLibrariesManager().get_box_data(library, name)

    # Init new box
    box_model = BoxModel(
        library,
        name,
        QPointF(box_reconstructor['geometry']['x'], box_reconstructor['geometry']['y']),
        QSizeF(box_reconstructor['geometry']['width'], box_reconstructor['geometry']['height']),
        box_reconstructor['geometry']['rotation'],
        True,
        box_reconstructor['text'],
        box_data.icon
    )

    # Build inputs groups
    box_model.add_inputs_groups(box_data.inputs, infinite_groups_reset=True)

    # Build outputs groups
    box_model.add_outputs_groups(box_data.outputs, infinite_groups_reset=True)

    for param_name in box_reconstructor['params']:
        box_model.create_parameter(param_name, box_reconstructor['params'][param_name]['type'], box_reconstructor['params'][param_name]['value'])

    # Common box informations (inputs, outputs)
    abstract_box_model(box_model, box_reconstructor)

    return box_model

def composite_box_input (input_reconstructor : Dict) -> BoxInputModel:
    """Permet de reconstruire une entrée de box composite suivant son reconstructeur."""
    raise Exception('No reconstructor created yet for the composite box input !!!')

def composite_box_output (output_reconstructor : Dict) -> BoxOutputModel:
    """Permet de reconstruire une sortie de box composite suivant son reconstructeur."""
    raise Exception('No reconstructor created yet for the composite box output !!!')

def abstract_box_model (box : AbstractBoxModel, box_reconstructor : Dict) -> AbstractBoxModel :
    """Permet de repeupler les données contenues dans le reconstructeur dans l'abstract box model."""

    # Inputs
    for inputs_group_dict in box_reconstructor['inputs_groups']:
        for input_index, input_dict in enumerate(inputs_group_dict['inputs']) :
            if inputs_group_dict["isInfinite"]:
                # Add input to group
                input_ = box.append_input(inputs_group_dict["name"])
            else:
                # Get input
                input_ = box.inputs_groups[inputs_group_dict["name"]].inputs[input_index]

            # Set input values
            input_.graphical_index = input_dict['graphical_index']
            input_.text = input_dict['text']

    # Outputs
    for outputs_group_dict in box_reconstructor['outputs_groups']:
        for output_index, output_dict in enumerate(outputs_group_dict['outputs']) :
            if outputs_group_dict["isInfinite"]:
                # Add output to group
                output = box.append_output(outputs_group_dict["name"])
            else:
                # Get output
                output = box.outputs_groups[outputs_group_dict["name"]].outputs[output_index]

            # Set output values
            output.graphical_index = output_dict['graphical_index']
            output.text = output_dict['text']

    abstract_graphical_element(box, box_reconstructor)
    return box

def abstract_graphical_element (graphical_element : AbstractGraphicalElement, element_reconstructor : Dict) -> AbstractGraphicalElement :
    """Permet de repeupler les données contenues dans le reconstructeur dans l'objet graphique."""
    graphical_element.graphical_index = element_reconstructor['graphical_index']
    return graphical_element

def decode (element_reconstructor : Dict) -> AbstractGraphicalElement :
    """Permet de retourner un l'élément contenu dans le dictionnaire."""

    if element_reconstructor is None :
        return

    if element_reconstructor['type'] == 'box' :
        return box(element_reconstructor)

    elif element_reconstructor['type'] == 'box-composite-input' :
        return composite_box_input(element_reconstructor)

    elif element_reconstructor['type'] == 'box-composite-output' :
        return composite_box_output(element_reconstructor)



#-----------------------------------
# Imports
#-----------------------------------

from typing import Any, Iterable, Union

from PyQt5 import QtGui
from PyQt5.QtCore import QPointF, QSizeF, Qt, pyqtSignal, QObject
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import QGraphicsItem

from muphyn.packages.core.plci_core_data_type import DataType
from muphyn.packages.core.utils.log_manager import LogManager
from muphyn.packages.interface.models.links_model.abstract_node_model import AbstractNodeModel
from muphyn.packages.interface.models.signals_model.abstract_signal_model import AbstractSignalModel
from muphyn.packages.interface.models.signals_model.signal_link_model import SignalLinkModel
from muphyn.packages.interface.shapes.basic_shape.abstract_shape import AbstractShape

#-----------------------------------
# Class
#-----------------------------------

static_io_size = QSizeF(10, 22)
static_io_size_2 = QSizeF(static_io_size.width() / 2, static_io_size.height() / 2)
static_io_path = QtGui.QPainterPath(QPointF(-static_io_size_2.width(), -static_io_size_2.height()))

static_io_path.lineTo(-static_io_size_2.width(), (-static_io_size_2.height()) + 1)
static_io_path.lineTo(static_io_size_2.width(), 0)
static_io_path.lineTo(-static_io_size_2.width(), static_io_size_2.height() + 1)
static_io_path.lineTo(-static_io_size_2.width(), static_io_size_2.height())

class AbstractConnectionModel (AbstractSignalModel, AbstractNodeModel[AbstractSignalModel]) : 
    """Est la classe abstraite communes aux models représentant des connexions entre les boxes et les signaux."""
    # -------------
    # Static Values
    # -------------
    ConnectorSide = 15
    ConnectorLineLength = 10

    ItemWidth = ConnectorSide + ConnectorLineLength
    ItemHeight = 2 * ConnectorSide - 3
    ItemSize = QSizeF(ItemWidth, ItemHeight)

    # -------------
    # Signals
    # -------------
    is_connected_changed = pyqtSignal()

    # -------------
    # Constructors
    # -------------

    def __init__ (self, name : str, data_type : DataType, position : QPointF, size : QSizeF, links : Iterable[SignalLinkModel] = [],
                  text : str = '', is_infinite : bool = False, parent : QGraphicsItem = None) :
        
        AbstractSignalModel.__init__(self, data_type)
        AbstractNodeModel.__init__(self, name, position, size, links, text, parent)

        self._connector: AbstractShape = None
        self._lbl: AbstractShape = None
        self._line: AbstractShape = None
        
        self._color = QColor(Qt.GlobalColor.black)

        self._is_infinite = is_infinite

        self.is_connected = self._links.__len__() > 0
        self.links_count_changed.connect(self.on_links_count_changed)

    # -------------
    # Properties
    # -------------
    @property
    def color(self) -> QColor:
        return self._color

    @color.setter
    def color(self, new_color: Union[QColor, Qt.GlobalColor]):
        if new_color != self._color:
            LogManager().debug("change color")
            if type(new_color) == Qt.GlobalColor:
                self._color = QColor(new_color)
            elif type(new_color) == QColor:
                self._color = new_color
            else:
                raise(TypeError(f"{self.__class__.name}.color: Wrong type provided {type(new_color)} instead of QColor or Qt.GlobalColor"))

            self._lbl._pen.setColor(self._color)
            self._line._pen.setColor(self._color)
            self._connector._pen.setColor(self._color)

    @property
    def is_infinite (self) -> bool :
        """Permet de savoir si la connexion est ou fait partit d'un nombre infinie de connexion."""
        return self._is_infinite

    @property
    def diagram_model (self) -> Any :
        """Permet de récuperer l'éditeur visuel."""
        # Get parent item
        parent = self.parent() if self.parent() is not None else self.parentItem()
        if parent is not None and hasattr(parent, "diagram_model"):
            return parent.diagram_model
        else:
            return

    @property
    def absolute_connector_center(self) -> QPointF:
        return self._connector.scenePos() + self._connector.boundingRect().center()

    @property
    def box_model(self):
        return self.parent().parent()

    # -------------
    # Methods
    # -------------
    def on_links_count_changed(self):
        is_connected = self._links.__len__() > 0
        if self.is_connected != is_connected:
            self.is_connected = is_connected
            self.is_connected_changed.emit()



#-----------------------------------
# Imports
#-----------------------------------

from PyQt5.QtWidgets import QWidget, QAction
from PyQt5.QtCore import pyqtSlot

#-----------------------------------
# Class
#-----------------------------------

class RecentFileMenu (QAction) :
    """Est le menu qui permet d'afficher et de gérer les fichiers récents."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, parent : QWidget, path : str, open_method : callable) :

        QAction.__init__(self, parent)
        self.setText(path)
        self.triggered.connect(self.click)
        self._path = path
        self._open_method = open_method

    # -------------
    # Properties
    # -------------

    @property
    def path (self) -> str :
        """Permet de récuperer le chemin pointé par le menu de fichier récent."""
        return self._path 

    @property
    def open_method (self) -> callable :
        """Permet de récuperer la méthode appelée pour ouvrir un fichier."""
        return self._open_method

    # -------------
    # Methods
    # -------------

    @pyqtSlot()
    def click (self) :
        """Est la méthode appelée lorsque l'utilisateur clique sur l'élément de fichier récent."""
        self._open_method(self._path)



#-----------------------------------
# Imports
#-----------------------------------

# General Imports
from typing import List

# PyQt5 Imports
from PyQt5.QtCore import QPointF, QSizeF, pyqtSignal, pyqtSlot
from PyQt5.QtWidgets import QGraphicsItem

# Project Imports
from muphyn.packages.interface.graphical_actions.diagram_move_graphical_element_action import DiagramMoveGraphicalElementAction
from muphyn.packages.interface.graphical_actions.diagram_resize_graphical_element_action import DiagramResizeGraphicalElementAction
from muphyn.packages.interface.graphical_actions.diagram_rotate_graphical_element_action import DiagramRotateGraphicalElementAction
from muphyn.packages.interface.models.event_signal_data import EventSignalData
from muphyn.packages.interface.models.graphical_models.abstract_graphical_element import AbstractGraphicalElement

#-----------------------------------
# Class
#-----------------------------------

class AbstractMoveableGraphicalElement (AbstractGraphicalElement) :
    """Est la classe abstraite commune aux éléments graphiques qui peuvent être bougé dans l'interface."""

    # -------------
    # Signals
    # -------------

    position_changed = pyqtSignal()
    size_changed = pyqtSignal()
    rotation_changed = pyqtSignal()

    # -------------
    # Constructors
    # -------------

    def __init__ (self, name : str, position : QPointF, size : QSizeF, rotation : float = 0.0, text : str = '', parent : QGraphicsItem = None) :
        
        AbstractGraphicalElement.__init__(self, name, position, rotation, text, parent)

        self._old_pos = position
        self._old_rot = rotation
        
        self.setFlag(QGraphicsItem.ItemIsMovable)
        self.setFlag(QGraphicsItem.ItemIsSelectable)
        self.setFlag(QGraphicsItem.ItemSendsScenePositionChanges)

        self.xChanged.connect(self.position_changed_method)
        self.yChanged.connect(self.position_changed_method)
        self.rotationChanged.connect(self.rotation_changed_method)
    
    # -------------
    # Methods
    # -------------
    def size_changed_method (self) -> None :
        self.size_changed.emit()

    def resize (self, param : EventSignalData) -> None :

        bounding_rect = self.boundingRect()

        old_size = self.size

        x2 = param.value.x() / 2
        y2 = param.value.y() / 2
        bounding_rect.adjust(-x2, -y2, x2, y2)

        self.prepareGeometryChange()
        # self.update()

        if bounding_rect.width() < 0 :
            bounding_rect.setWidth(0)
    
        if bounding_rect.height() < 0 :
            bounding_rect.setHeight(0)

        #UNDO REDO Stuff

        if self.action_size_semaphore :
            return

        if not(hasattr(self, 'diagram_model')) :
            return

        if self.diagram_model is None : 
            return

        if self.diagram_model.actions_holder is None :
            return
        
        last_action = self.diagram_model.actions_holder.last_action
        if not(last_action is None) :
            if isinstance(last_action, DiagramResizeGraphicalElementAction) :
                if last_action.graphical_index == self.graphical_index :
                    last_action.new_size = self.size
                    return
        
        self.diagram_model.actions_holder.append(DiagramResizeGraphicalElementAction(self, old_size, self.size))

    # -------------
    # Slots
    # -------------
    @pyqtSlot()
    def position_changed_method (self) -> None : 
        self.position_changed.emit()

        #UNDO REDO Stuff

        new_pos = self.pos()

        if self._old_pos == new_pos :
            return

        if self.action_pos_semaphore :
            return

        if not(hasattr(self, 'diagram_model')) :
            return

        if self.diagram_model is None : 
            return

        if self.diagram_model.actions_holder is None :
            return
        
        last_action = self.diagram_model.actions_holder.last_action

        if not(last_action is None) :
            if isinstance(last_action, DiagramMoveGraphicalElementAction) :

                if self.graphical_index == last_action._graphical_element_index :
                    last_action.new_position = new_pos
                    self._old_pos = new_pos
                    return

                if last_action.contains_index(self.graphical_index) :
                    return

        selection : List[AbstractGraphicalElement] = []

        for selected_element in self.diagram_model._scene.selectedItems() :
            selection.append(selected_element)

        self.diagram_model.actions_holder.append(DiagramMoveGraphicalElementAction(self, selection, self._old_pos, new_pos))
        self._old_pos = new_pos

    @pyqtSlot()
    def rotation_changed_method (self) -> None :
        self.rotation_changed.emit()

        new_rot = self.rotation()

        #UNDO REDO Stuff

        if self._old_rot == new_rot :
            return

        if self.action_rot_semaphore :
            return

        if not(hasattr(self, 'diagram_model')) :
            return

        if self.diagram_model is None : 
            return

        if self.diagram_model.actions_holder is None :
            return
        
        last_action = self.diagram_model.actions_holder.last_action

        if not(last_action is None) :
            if isinstance(last_action, DiagramRotateGraphicalElementAction) :
                if last_action.graphical_index == self.graphical_index :
                    last_action.new_rotate = new_rot
                    self._old_rot = new_rot
                    return
        
        self.diagram_model.actions_holder.append(DiagramRotateGraphicalElementAction(self, self._old_rot, new_rot))
        self._old_rot = new_rot
    


#-----------------------------------
# Imports
#-----------------------------------

from datetime import date
from typing import Iterable, List, Any, Dict

from muphyn.packages.interface.models.editable_models.abstract_editable_model import AbstractEditableModel
from muphyn.packages.interface.models.graphical_models.box_input_model import BoxInputModel
from muphyn.packages.interface.models.graphical_models.box_output_model import BoxOutputModel

#-----------------------------------
# Class
#-----------------------------------

class AbstractBoxModel (AbstractEditableModel) : 
    """Est la classe abstraite commune aux classes capables de garder les informations de bases des boxes."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, name : str, path : str, creator : str, date_creation : date, version : float, inputs : Iterable[BoxInputModel] = [], outputs : Iterable[BoxOutputModel] = [], properties : Dict[str, Any] = {}) :
        
        AbstractEditableModel.__init__(self, name, path, creator, date_creation, version)

        self._inputs : List[BoxInputModel] = list(inputs)
        self._outputs : List[BoxOutputModel] = list(outputs)
        self._properties : Dict[str, Any] = {}

    # -------------
    # Properties
    # -------------

    @property
    def inputs (self) -> Iterable[BoxInputModel] :
        """Permet de récuperer la liste des entrées."""
        return self._inputs

    @property
    def outputs (self) -> Iterable[BoxOutputModel] :
        """Permet de récuperer la liste des sorties."""
        return self._outputs



from muphyn.packages.core.plci_core_scheduler import Scheduler

from muphyn.packages.core.plci_core_data_type import DataType
from muphyn.packages.core.plci_core_signal import Signal
from muphyn.packages.core.plci_core_scheduler_params import SchedulerParams
from muphyn.packages.core.plci_core_diagram import Diagram

from muphyn.packages.unit_test.unit_test import get_boxes_libraries, get_schedulers_library, simulate

# Get library
boxesLibraries = get_boxes_libraries()
schedulersLibraries = get_schedulers_library()

# Scheduler parameters
scheduler : Scheduler = schedulersLibraries.construct_scheduler('Schedulers', 'Default')
scheduler.params = SchedulerParams(5.0, 0.001)

## DIAGRAM 1
diagram1 : Diagram = Diagram()
diagram1.append(Signal(index_ = 0, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
diagram1.append(Signal(index_ = 1, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
diagram1.append(Signal(index_ = 2, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
diagram1.append(Signal(index_ = 3, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
diagram1.append(Signal(index_ = 4, signal_type_ = DataType.FLOAT, default_value_ = 0.0))

diagram1.append(boxesLibraries.construct_box('Boxes.Math', 'Addition'))
diagram1.append(boxesLibraries.construct_box('Boxes.Sources', 'Constant', value = 3))
diagram1.append(boxesLibraries.construct_box('Boxes.Signals', 'Amplifier', gain = 10))

diagram1.add_box_inputs(diagram1.boxes[0], diagram1.signals[0], diagram1.signals[1], diagram1.signals[2])
diagram1.add_box_inputs(diagram1.boxes[2], diagram1.signals[3])

diagram1.add_box_outputs(diagram1.boxes[0], diagram1.signals[3])
diagram1.add_box_outputs(diagram1.boxes[1], diagram1.signals[2])
diagram1.add_box_outputs(diagram1.boxes[2], diagram1.signals[4])

## DIAGRAM 2
scheduler.diagram.append(Signal(index_ = 0, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
scheduler.diagram.append(Signal(index_ = 1, signal_type_ = DataType.FLOAT, default_value_ = 0.0))
scheduler.diagram.append(Signal(index_ = 2, signal_type_ = DataType.FLOAT, default_value_ = 0.0))

scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sources', 'Step', start_value = 0, stop_value = 1, step_time = 0.5))
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sources', 'Ramp', start_time = 0, initial_value = 0, slope = 0.1))
scheduler.diagram.append(boxesLibraries.construct_box('Boxes.Sinks', 'Graph', start_time = 0, stop_time = scheduler.stop_time))

scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[2], scheduler.diagram.signals[2])

scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[0], scheduler.diagram.signals[0])
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[1], scheduler.diagram.signals[1])

print("Diagram 1 boxes : ")
for box in diagram1.boxes :
    print("\t - ", box)

print("Diagram 2 boxes : ")
for box in scheduler.diagram.boxes :
    print("\t - ", box)


scheduler.diagram.append(diagram1)
scheduler.diagram.add_linked_signals(scheduler.diagram.signals[0], scheduler.diagram.signals[3])
scheduler.diagram.add_linked_signals(scheduler.diagram.signals[1], scheduler.diagram.signals[4])
scheduler.diagram.add_linked_signals(scheduler.diagram.signals[7], scheduler.diagram.signals[2])

# Launch solver
simulate(scheduler)


#-----------------------------------
# Imports
#-----------------------------------

from typing import List
from numpy import sin

#-----------------------------------
# Functions
#-----------------------------------

def _init_saw_tooth_box (box, simulation_params) -> None :

    if not 'amplitude' in box :
        box['amplitude'] = 1.0

    if not 'slope' in box :
        box['slope'] = 1.0
    
    if not 'mean_value' in box :
        box['mean_value'] = 0.0

    box['period'] = box['amplitude'] / box['slope']
    if box['period'] < 0 :
        box['period'] = - box['period']

    box['current_period'] = 0.0
    box['start_value'] = box['mean_value'] - box['amplitude'] / 2

def _function_saw_tooth_box (box, event_) -> List : 
    
    if event_.timing == 0 :
        box['current_period'] = 0.0
    else :
        box['current_period'] = box['current_period'] + event_.step_time

    if box['current_period'] > box['period'] :
        box['current_period'] = box['current_period'] - box['period']

    v = box['start_value'] + box['current_period'] * box['slope']
    
    events : List = []

    for output in box.outputs :
        events.append(box.construct_signal_event(output, v))

    return events


#-----------------------------------
# Imports
#-----------------------------------

from typing import List, Any

from PyQt5.QtWidgets import QTableWidget, QTableWidgetItem, QComboBox

from muphyn.packages.interface.graphical_actions.abstract_graphical_action import AbstractGraphicalAction
from muphyn.packages.core.plci_core_data_type import DataType

#-----------------------------------
# Class
#-----------------------------------

class NewBoxDialogAddAction (AbstractGraphicalAction) :
    """Est l'action capable d'ajouter un item dans un des tableau."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, source : Any, text : str, table : QTableWidget) :
        AbstractGraphicalAction.__init__(self, source)
        self._table = table
        self._current_index = -1
        self._was_already_added = False
        self._old_data : List[str] = []
        self._text : str = text

    # -------------
    # Properties
    # -------------

    @property
    def current_index (self) -> int :
        """Permet de récuperer l'index au moment où la nouvelle valeur a été ajouté."""
        return self._current_index

    @property 
    def table (self) -> QTableWidget :
        """Permet de récuperer le tableau dans lequel l'item a été ajouté."""
        return self._table

    @property
    def was_already_added (self) -> bool :
        """
        Permet de savoir si c'est la deuxième fois que le do est appelée.
        Permet de rajouter une ligne de données contenant les valeurs avant suppression.
        """
        return self._was_already_added

    @property
    def old_data (self) -> List[str] :
        """Permet de récuperer les données qui se trouvait dans le tableau avant le undo."""
        return self._old_data
    
    # -------------
    # Method
    # -------------

    def do (self) -> None :

        self._current_index = self._table.rowCount()
        self._table.setRowCount(self._table.rowCount() + 1)
        self._table.setRowHeight(self._current_index, 20)

        if self._was_already_added :
            self._table.setItem(self._current_index, 0, QTableWidgetItem(self._old_data[0]))
            
            combo_type = QComboBox()
            for data_type in DataType :
                combo_type.addItem(data_type.__str__().lower())
            combo_type.setCurrentText(self._old_data[1])

            self._table.setCellWidget(self._current_index, 1, combo_type)

            combo_input_count = QComboBox()
            for i in range(9) :
                combo_input_count.addItem(str(i + 1))
            combo_input_count.addItem('Infinity')
            combo_input_count.setCurrentText(self._old_data[2])

            self._table.setCellWidget(self._current_index, 2, combo_input_count)

            self._table.setItem(self._current_index, 3, QTableWidgetItem(self._old_data[3]))

        else :
            self._table.setItem(self._current_index, 0, QTableWidgetItem(self._text + ' ' + str(self._current_index + 1)))

            combo_type = QComboBox()
            for data_type in DataType :
                combo_type.addItem(data_type.__str__().lower())
            combo_type.setCurrentText('float')

            self._table.setCellWidget(self._current_index, 1, combo_type)

            combo_input_count = QComboBox()
            for i in range(9) :
                combo_input_count.addItem(str(i + 1))
            combo_input_count.addItem('Infinity')

            self._table.setCellWidget(self._current_index, 2, combo_input_count)
            self._table.setItem(self._current_index, 3, QTableWidgetItem('0'))

    def undo (self) -> None :

        self._was_already_added = True

        self._old_data.append(self._table.item(self._current_index, 0).text())
        self._old_data.append(self._table.cellWidget(self._current_index, 1).currentText())
        self._old_data.append(self._table.cellWidget(self._current_index, 2).currentText())
        self._old_data.append(self._table.item(self._current_index, 3).text())

        self._table.removeRow(self._current_index)


from PyQt5 import QtCore, QtGui, QtWidgets


class Diedrico (QtWidgets.QWidget) :
    
    def paintEvent(self, event):
        qp = QtGui.QPainter(self)
        pen = QtGui.QPen(QtGui.QColor(QtCore.Qt.black), 5)
        qp.setPen(pen)
        qp.drawRect(500, 500, 1000, 1000)
        qp.drawRect(100, 100, 10, 20)
        qp.drawText(100, 100, "Test")


class UiVentana(QtWidgets.QMainWindow):
    factor = 1.5

    def __init__(self, parent=None):
        super(UiVentana, self).__init__(parent)

        self._scene = QtWidgets.QGraphicsScene(self)
        self._view = QtWidgets.QGraphicsView(self._scene)

        self._diedrico = Diedrico()
        self._diedrico.setFixedSize(2000, 2000)
        self._scene.addWidget(self._diedrico)

        self.setCentralWidget(self._view)

        QtWidgets.QShortcut(
            QtGui.QKeySequence(QtCore.Qt.Key_Plus), # <---
            self._view,
            context=QtCore.Qt.WidgetShortcut,
            activated=self.zoom_in,
        )

        QtWidgets.QShortcut(
            QtGui.QKeySequence(QtCore.Qt.Key_Minus), # <---
            self._view,
            context=QtCore.Qt.WidgetShortcut,
            activated=self.zoom_out,
        )

    @QtCore.pyqtSlot()
    def zoom_in (self):
        scale_tr = QtGui.QTransform()
        scale_tr.scale(UiVentana.factor, UiVentana.factor)

        tr = self._view.transform() * scale_tr
        self._view.setTransform(tr)

    @QtCore.pyqtSlot()
    def zoom_out (self):
        scale_tr = QtGui.QTransform()
        scale_tr.scale(UiVentana.factor, UiVentana.factor)

        scale_inverted, invertible = scale_tr.inverted()

        if invertible:
            tr = self._view.transform() * scale_inverted
            self._view.setTransform(tr)


if __name__ == "__main__":
    import sys

    app = QtWidgets.QApplication(sys.argv)
    ui = UiVentana()
    ui.show()
    sys.exit(app.exec_())


from muphyn.packages.core.simulation.model_parser.open_modelica_model_parser import OpenModelicaModelParser, OpenModelicaComponent, OpenModelicaComponentAnnotation, OpenModelicaConnection, OpenModelicaConnectionAnnotation, OpenModelicaInput, OpenModelicaLine, OpenModelicaOutput, OpenModelicaTransformation


#-----------------------------------
# Imports
#-----------------------------------

from muphyn.packages.core.plci_core_scheduler_params import SchedulerParams

#-----------------------------------
# Class
#-----------------------------------

class SchedulerModel :
    """Est le model de données pour l'éditeur des planificateurs."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, library : str, name : str, params : SchedulerParams) :

        self._library : str = library
        self._name : str = name
        self._params = params

    # -------------
    # Properties
    # -------------
    @property
    def completeName(self) -> str:
        return f"{self._library}.{self._name}"
    
    @property
    def library (self) -> str :
        """Permet de récuperer le bibliothèque dans laquelle se trouve le planificateur."""
        return self._library
    
    @library.setter 
    def library (self, library_ : str) -> None :
        """Permet de modifier la bibliothèque dans laquelle se trouve le planificateur."""
        self._library : str = library_

    @property
    def name (self) -> str :
        """Permet de récuperer le nom du planificateur."""
        return self._name
    
    @name.setter
    def name (self, name_ : str) -> None :
        """Permet de modifier le nom du planificateur."""
        self._name = name_

    @property
    def params (self) -> SchedulerParams :
        """Permet de récuperer les paramètres du planificateur."""
        return self._params
    
    @params.setter
    def params (self, params_ : SchedulerParams) -> None :
        """Permet de modifier les paramètres du planificateur."""
        self._params = params_ 


#-----------------------------------
# Imports
#-----------------------------------

# General Imports
import yaml
from typing import Any, Dict

# PyQt5 Imports
from PyQt5.QtCore import QPointF

# Project Imports
from muphyn.packages.interface.graphical_actions.abstract_diagram_action import AbstractDiagramAction
from muphyn.packages.interface.models.graphical_models.abstract_box_model import AbstractBoxModel
from muphyn.packages.interface.models.graphical_models.abstract_graphical_element import AbstractGraphicalElement
from muphyn.packages.interface.models.signals_model.signal_link_model import SignalLinkModel
from muphyn.packages.interface.models.graphical_models.box_model import BoxModel

#-----------------------------------
# Class
#-----------------------------------

class DiagramPasteGraphicalElementAction (AbstractDiagramAction) :
    """Est l'action capable de coller le contenus d'une box."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, diagram_editor, diagram_model : Any, content_paste : str) :
        
        AbstractDiagramAction.__init__(self, diagram_model)

        self._serialized_diagram = None
        self._diagram_editor = diagram_editor
        self._content_paste = content_paste
        self._elements_indices = []

    # -------------
    # Methods
    # -------------

    def set_or_get_graphical_index (self, age : AbstractGraphicalElement, dictionary : Dict) -> None :
        """Permet de récuperer ou de sélectionner le graphical index pour l'élément passé en paramètre."""
        
        if 'graphical_index' in dictionary : 
            age.graphical_index = dictionary['graphical_index']
        
        else :
            dictionary['graphical_index'] = age.graphical_index
            self._should_reload = True

            if isinstance(age, SignalLinkModel) or isinstance(age, AbstractBoxModel) : 
                if not(age.graphical_index in self._elements_indices) :
                    self._elements_indices.append(age.graphical_index)

    def do (self) :
        
        self._diagram_editor.unslect_elements()

        if self._serialized_diagram is None : 
            self._serialized_diagram = yaml.load(self._content_paste, yaml.FullLoader)

        if 'MuPhyN' in self._serialized_diagram :
            
            signals = self._serialized_diagram['MuPhyN']['signals']
            
            for signal in signals : 
                signal['input'] = None
                signal['output'] = None

            for box_dict in self._serialized_diagram['MuPhyN']['boxes'] :
                # Build new Box Model from data dict
                box_model = BoxModel.fromDict(box_dict)

                # Move new Box Model
                box_model.setPos(box_model.pos() + QPointF(box_model.size.width()/2, box_model.size.height()/2))

                # Get Graphical Index of new box
                self.set_or_get_graphical_index(box_model, box_dict)

                # Add new Box Model to diagram editor
                self._diagram_model.add_element(box_model)

                # Set new Box Model as selected
                box_model.setSelected(True)

    def undo (self) :
        
        for element_index in self._elements_indices :

            element = self.diagram_model.get_element_by_graphical_index(element_index)

            if element is None : 
                continue

            if isinstance(element, SignalLinkModel) :
                element.unbind()
                self.diagram_model.remove_element(element)

        for element_index in self._elements_indices :

            element = self.diagram_model.get_element_by_graphical_index(element_index)

            if element is None : 
                continue

            if not(isinstance(element, SignalLinkModel)) :
                self.diagram_model.remove_element(element)


#-----------------------------------
# Imports
#-----------------------------------

from typing import Iterable
from PyQt5 import QtGui

from PyQt5.QtCore import QPointF, QRectF, QSizeF
from PyQt5.QtWidgets import QGraphicsItem, QStyleOptionGraphicsItem

from muphyn.packages.interface.models.links_model.abstract_node_model import AbstractNodeModel
from muphyn.packages.interface.models.links_model.link_model import LinkModel

#-----------------------------------
# Class
#-----------------------------------

class NodeModel (AbstractNodeModel[LinkModel]) :
    """Est le type générique de noeud pour afficher des liens non typés dans l'interface."""

    connector_size: QSizeF = QSizeF(20, 20)
    line_length: float = 20.0

    node_size: QSizeF = QSizeF(connector_size.x(), connector_size.y()+line_length)
    
    # -------------
    # Constructors
    # -------------
    
    def __init__ (self, position : QPointF, size : QSizeF, links : Iterable[LinkModel] = [], text : str = '', parent : QGraphicsItem = None) :

        AbstractNodeModel.__init__(self, 'node', position, size, links, text, parent)

    # -------------
    # Methods
    # -------------

    def insert_link (self, index : int, link : LinkModel) -> None :
        """Permet d'insérer un lien à la position données."""

        if link is None : 
            return

        if index < 0 :
            return

        if index > self._links.__len__() :
            return

        self._links.insert(index, link)

    def remove_link (self, link : LinkModel) -> None :
        """Permet de supprimer un lien."""
        self._links.remove(link)
    
    def paint(self, painter: QtGui.QPainter, option: QStyleOptionGraphicsItem, widget) -> None:
        painter.drawEllipse(QRectF(self.position, NodeModel.connector_size))
        painter.drawLine(
            QPointF(NodeModel.width(), NodeModel.height()/2), 
            QPointF(NodeModel.width() + NodeModel.line_length, NodeModel.height()/2)
        )
        super().paint(painter, option, widget)


#-----------------------------------
# Imports
#-----------------------------------

from typing import List

#-----------------------------------
# Functions
#-----------------------------------

def _init_integrator_box (box, simulation_params) -> None :
    box['last_output'] = 0
    box['last_input'] = 0
    box['last_timing'] = 0


def _function_integrator_box (box, event_) -> List:
        
    input = box.get_input(0).value

    if event_.timing == 0 :
        return []

    if box['last_timing'] == event_.timing :
        return []

    box['last_timing'] = event_.timing

    v = (event_.step_time * (input + box['last_input']) / 2) + box['last_output']

    box['last_output'] = v
    box['last_input'] = input

    events : List = []
    
    for output in box.outputs :
        events.append(box.construct_signal_event(output, v))

    return events


"""
This file is deprecated - MUST BE DELETED
"""

#-----------------------------------
# Class
#-----------------------------------

class Color :
    """Est le modèle de données pour les couleurs."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, red : float, green : float, blue : float) :

        self._red = red
        self._green = green
        self._blue = blue

    # -------------
    # Properties
    # -------------

    @property
    def red (self) -> float :
        """Permet de récuperer la valeur rouge de la couleur."""
        return self._red

    @property
    def green (self) -> float : 
        """Permet de récuperer la valeur verte de la couleur."""
        return self._green

    @property
    def blue (self) -> float : 
        """Permet de récuperer la valeur bleue de la couleur."""
        return self._blue

#-----------------------------------
# Static Elements
#-----------------------------------

black : Color = Color(0, 0,  0)
white : Color = Color(1, 1, 1)
gray_25 : Color = Color(0.25, 0.25, 0.25)
gray_50 : Color = Color(0.50, 0.50, 0.50)
gray_75 : Color = Color(0.75, 0.75, 0.75)
red : Color = Color(1, 0, 0)
green : Color = Color(0, 1, 0)
blue : Color = Color(0, 0, 1)
yellow : Color = Color(1, 1, 0)
cyan : Color = Color(0, 1, 1)
magenta : Color = Color(1, 0, 1)



#-----------------------------------
# Imports
#-----------------------------------

from datetime import date

from muphyn.packages.interface.models.model_signal import ModelSignal

#-----------------------------------
# Class
#-----------------------------------

class AbstractEditableModel :
    """Est la classe abstraite commune aux éléments affichable et modifiable dans l'interface graphique."""

    # -------------
    # Signals
    # -------------

    # -------------
    # Constructors
    # -------------

    def __init__ (self, name : str, path : str, creator : str, date_creation : date, version : float) :

        self._name : str = name
        self._path : str = path
        self._is_unsaved : bool = False

        self.creator = creator
        self.date_creation = date_creation
        self.version = version

        self.name_changed = ModelSignal()
        """Est le signal utilisé quand le modèle voit son nom changer."""

        self.saved_changed = ModelSignal()
        """Est le signal utilisé quand le modèle est sauvegarder."""

        self.selection_changed = ModelSignal()
        """Est le signal utilisé quand le modèle voit sa sélection d'élément modifié."""

    # -------------
    # Properties
    # -------------

    @property
    def name (self) -> str :
        """Permet de récuperer le nom de l'élément mdoifiable."""
        return self._name
    
    @name.setter
    def name (self, name_ : str) -> None :
        """Permet de modifier le nom de l'élément modifiable."""
        self._name : str = name_.__str__()
        self.name_changed.emit(name = self._name)

    @property
    def path (self) -> str :
        """Permet de récuperer le chemin d'accès vers le fichier contenant l'élément modifiable."""
        return self._path

    @path.setter
    def path (self, path_ : str) -> None :
        """Permet de modifier le chemin d'accès vers el fichier contenant l'élément modifiable."""
        self._path : str = path_.__str__()

    @property
    def is_unsaved (self) -> bool :
        """Permet de récuperer l'état de sauvegarde actuel de l'élement modifiable."""
        return self._is_unsaved

    @is_unsaved.setter
    def is_unsaved (self, is_unsaved_ : bool) -> None :
        """Permet de modifier l'état de sauvegarde actuel de l'élément modifiable."""
        self._is_unsaved : bool = is_unsaved_
        self.saved_changed.emit()

    @property
    def creator (self) -> str :
        """Permet de récuperer le nom du créateur."""
        return self._creator

    @creator.setter
    def creator (self, creator_ : str) -> None :
        """Permet de modifier le nom créateur."""
        self._creator : str = creator_ or ''

    @property 
    def date_creation (self) -> date :
        """Permet de récuperer la date à laquelle l'élément a été créé."""
        return self._date_creation

    @date_creation.setter 
    def date_creation (self, date_creation_ : date) -> None : 
        """Permet de modifier la date à laquelle l'élément a été créé."""
        self._date_creation : date = date_creation_

    @property
    def version (self) -> float :
        """Permet de récuperer la version actuelle de l'élément."""
        return self._version

    @version.setter
    def version (self, version_ : float) -> None :
        """Permet de modifier la version actuelle de l'élément."""

        if version_ < 0 :
            return

        self._version : float = version_
        
    # -------------
    # Methods
    # -------------

    def set_path_name (self, path : str = None, name : str = None) -> None :
        """Permet de sauvegarder le modèle dans le dossier/fichier renseigné."""
        self.path = path
        self.name = name


#-----------------------------------
# Imports
#-----------------------------------

from typing import Any

from muphyn.packages.interface.models.links_model.abstract_link_model import AbstractLinkModel
from muphyn.packages.interface.models.links_model.link_type import LinkType

#-----------------------------------
# Class
#-----------------------------------

class LinkModel (AbstractLinkModel) :
    """Est le type générique de lien pour afficher des liens non typés dans l'interface."""
    
    # -------------
    # Constructors
    # -------------

    def __init__ (self, node_model_1 : Any, node_model_2 : Any, link_type : LinkType,
                  text : str = '') :

        AbstractLinkModel.__init__(self, node_model_1, node_model_2, link_type, text)

    # -------------
    # Methods
    # -------------
    


#-----------------------------------
# Imports
#-----------------------------------

from typing import List, Any

from PyQt5.QtWidgets import QTableWidget, QTableWidgetItem, QComboBox

from muphyn.packages.interface.graphical_actions.abstract_graphical_action import AbstractGraphicalAction
from muphyn.packages.core.plci_core_data_type import DataType

#-----------------------------------
# Class
#-----------------------------------

class NewBoxDialogRemoveAction (AbstractGraphicalAction) :
    """Est l'action capable de supprimer des items dans un des tableau."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, source : Any, table : QTableWidget) :
        AbstractGraphicalAction.__init__(self, source)
        self._table = table
        self._data : List[List[str]] = []
        self._was_already_removed : bool = False

    # -------------
    # Properties
    # -------------

    @property 
    def table (self) -> QTableWidget :
        """Permet de récuperer le tableau dans lequel l'item a été ajouté."""
        return self._table

    @property
    def data (self) -> List[List[str]] :
        """Permet de récuperer les données qui ont été supprimées dans le but de les rajouter dans le tableau."""
        return self._data

    @property
    def was_already_removed (self) -> bool :
        """
        Permet de savoir si c'est la deuxième fois que le do est appelée.
        Permet de supprimer les lignes de données contenant les valeurs avant rajout.
        """
        return self._was_already_removed
    
    # -------------
    # Method
    # -------------

    def do (self) -> None :
        
        if self._was_already_removed :
            
            self._data.reverse()

            for row_data in self._data :

                row_data[1] = self._table.item(row_data[0], 0).text()
                row_data[2] = self._table.cellWidget(row_data[0], 1).currentText()
                row_data[3] = self._table.cellWidget(row_data[0], 2).currentText()
                row_data[4] = self._table.item(row_data[0], 3).text()

                self._table.removeRow(row_data[0])
                
        else :

            for d in self._data :
                d.clear()
            self._data.clear()

            rows : List[int] = []
            for i in self._table.selectedIndexes() :
                rows.append(i.row())

            rows.sort()
            rows.reverse()

            for row in rows :
                current_data : List[str] = []

                current_data.append(row)
                current_data.append(self._table.item(row, 0).text())
                current_data.append(self._table.cellWidget(row, 1).currentText())
                current_data.append(self._table.cellWidget(row, 2).currentText())
                current_data.append(self._table.item(row, 3).text())

                self._data.append(current_data)
                self._table.removeRow(row)
                
        self._was_already_removed = True
            
    def undo (self) -> None :

        self._data.reverse()
        for row_data in self._data :
            self._table.insertRow(row_data[0])
            self._table.setItem(row_data[0], 0, QTableWidgetItem(row_data[1]))

            combo_type = QComboBox()
            for data_type in DataType :
                combo_type.addItem(data_type.__str__().lower())
            combo_type.setCurrentText(row_data[2])

            self._table.setCellWidget(row_data[0], 1, combo_type)

            combo_input_count = QComboBox()
            for i in range(9) :
                combo_input_count.addItem(str(i + 1))
            combo_input_count.addItem('Infinity')
            combo_input_count.setCurrentText(row_data[3])

            self._table.setCellWidget(row_data[0], 2, combo_input_count)
            self._table.setItem(row_data[0], 3, QTableWidgetItem(row_data[4]))

            self._table.setRowHeight(row_data[0], 20)


#-----------------------------------
# Imports
#-----------------------------------

# General Imports
from datetime import date
from typing import Iterable

# Project Imports
from muphyn.packages.interface.models.graphical_models.abstract_box_model import AbstractBoxModel
from muphyn.packages.interface.models.editable_models.abstract_diagram_model import AbstractDiagramModel
from muphyn.packages.interface.models.graphical_models.abstract_graphical_element import AbstractGraphicalElement
from muphyn.packages.interface.models.graphical_models.box_input_model import BoxInputModel
from muphyn.packages.interface.models.graphical_models.box_output_model import BoxOutputModel

#-----------------------------------
# Class
#-----------------------------------

class BoxCompositeModel (AbstractBoxModel, AbstractDiagramModel) :
    """Est le modèle pour l'éditeur de box composite."""

    # -------------
    # Constructors
    # -------------

    def __init__(self, name : str, path : str, creator : str, date : date, version : float, inputs : Iterable[BoxInputModel] = [], outputs : Iterable[BoxOutputModel] = [], graphical_elements : Iterable[AbstractGraphicalElement] = []) :
        
        AbstractBoxModel.__init__(self, name, path, creator, date, version, inputs, outputs)
        AbstractDiagramModel.__init__(self, graphical_elements)


#-----------------------------------
# Imports
#-----------------------------------

from typing import Callable, Any
from ..plci_core_scheduler import Scheduler
from ..plci_core_scheduler_exception import SchedulerException

#-----------------------------------
# Class
#-----------------------------------

class SchedulerData : 
    """Est la classe commune des classes de données de bibliothèque."""

    # -------------
    # Constructors
    # -------------
    def __init__ (
            self, 
            scheduler_name : str, 
            scheduler_library : str,
            scheduler_method : Callable[[Scheduler], SchedulerException],
            creator : str,
            date_created : Any,
            version : float
        ) :
        
        self._scheduler_name : str = scheduler_name
        self._scheduler_library : str = scheduler_library
        self._scheduler_method : Callable[[Scheduler], SchedulerException] = scheduler_method
        self._creator = creator
        self._date_created = date_created
        self._version = version

    # -------------
    # Properties
    # -------------

    @property 
    def scheduler_name (self) -> str :
        """Permet de récuperer le nom du planificateur."""
        return self._scheduler_name

    @property
    def scheduler_library (self) -> str :
        """Pemet de récuperer la bibliothèque du planificateur."""
        return self._scheduler_library

    @property
    def scheduler_method (self) -> Callable[[Scheduler], SchedulerException] :
        """Permet de récuperer la méthode du planificateur."""
        return self._scheduler_method

    @property
    def creator (self) -> str :
        """Permet de récuperer le nom de la personne qui a créé le planificateur."""
        return self._creator

    @property
    def date_created (self) -> Any :
        """Permet de récuperer la date à laquelle le planificateur a été créée."""
        return self._date_created

    @property
    def version (self) -> float :
        """Permet de récuperer la version du planificateur."""
        return self.version

    # -------------
    # Methods
    # -------------
    
    def construct_scheduler (self) -> Scheduler :
        """Permet de générer un planificateur."""
        return Scheduler(self._scheduler_name, self._scheduler_method)

    def __str__ (self) -> str :
        """Permet de récuperer l'importeur de planificateur sous la forme de texte."""
        return self._scheduler_library + '.' + self._scheduler_name


#-----------------------------------
# Imports
#-----------------------------------

# PyQt5 Imports
from PyQt5.QtGui import QColor
from PyQt5.QtCore import Qt

# Project Imports
from muphyn.packages.core.plci_core_data_type import DataType
from muphyn.packages.interface.graphical_actions.abstract_unique_element_diagram_action import AbstractUniqueElementDiagramAction
from muphyn.packages.interface.models.graphical_models.abstract_graphical_element import AbstractGraphicalElement

#-----------------------------------
# Class
#-----------------------------------

class DiagramAddOutputToBoxAction (AbstractUniqueElementDiagramAction) :
    """Est la classe d'action permettant d'ajouter une sortie à une box."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, graphical_element : AbstractGraphicalElement, output_group_name : str, output_index: int = -1) :

        AbstractUniqueElementDiagramAction.__init__(self, graphical_element)

        self._output_added_graphical_index = None
        self._output_group_name = output_group_name
        self._output_index = output_index

    # -------------
    # Methods
    # -------------

    def do (self) :
        
        # Append output
        if self._output_index == -1:
            output_model = self.graphical_element.append_output(self._output_group_name)
        else:
            output_model = self.graphical_element.insert_output(self._output_group_name, self._output_index)
        
        if self._output_added_graphical_index is None :
            # Get the output graphical index if not already added before
            self._output_added_graphical_index = output_model.graphical_index

        else :
            # Set the output graphical index if already added before
            output_model.graphical_index = self._output_added_graphical_index


    def undo (self) :
        
        box = self.graphical_element

        if box is None : 
            return

        for output_index in range(box.output_len - 1, -1, -1) :

            output = box.get_output(output_index)
            if output.graphical_index == self._output_added_graphical_index :             
                box.remove_output(output)


# General Imports
import numbers
import os
import re
from typing import Any
from threading import Thread

# OPCUA connection
from opcua import Client, Node
from opcua.ua.uatypes import LocalizedText

# Open Modelica Libraries Imports
from OMPython import ModelicaSystem

from muphyn.packages.core.simulation.abstract_simulation import AbstractSimulation
from muphyn.packages.core.utils.directory_manager import DirectoryManager
from muphyn.packages.core.utils.log_manager import LogManager


class OpenModelicaSimulationController:
    def __init__(self, url: str):
        # URL
        self._url = url

        # OPC-UA client
        self._client = Client(url)

    # --------------
    # Properties
    # --------------
    @property
    def url(self) -> str:
        return self._url

    @url.setter
    def url(self, new_url) -> None:
        if self._url != new_url:
            self._url = new_url

            # Reset client
            self._client.disconnect()
            self._client = Client(new_url)
        
    # --------------
    # Methods
    # --------------
    def init_connection(self):
        try:
            # Connect to the client
            self._client.connect()

            # Get object nodes
            self._object_nodes = self.get_object_nodes()

            # Get run node
            self._run_node = self.get_node("run")

            # Get step node
            self._step_node = self.get_node("step")

            # Get realTimeScalingFactor node
            self._real_time_scaling_factor = self.get_node("realTimeScalingFactor")

            # Get enableStopTime node
            self._enable_stop_time = self.get_node("enableStopTime")

            # Get time node
            self._time = self.get_node("time")

            # Get terminate node
            self._terminate = self.get_node("terminate")
        except: 
            LogManager().error("Connection not established")

    def disconnect(self):
        self._client.disconnect()

    def get_object_nodes(self) -> dict:
        # Init object node dict
        object_nodes = {}

        # Get root node
        root_node: Node = self._client.get_root_node()
        root_node_display_name: LocalizedText = root_node.get_display_name()

        for child_node in root_node.get_children():
            root_child_display_name: LocalizedText = child_node.get_display_name()
            if root_child_display_name.Text == "Objects":
                for object_node in child_node.get_children():
                    object_node: Node = object_node
                    object_node_display_name = object_node.get_display_name()
                    object_nodes[object_node_display_name.Text] = object_node

        return object_nodes

    def get_node(self, node_name: str) -> Node:
        return self._object_nodes[node_name]

    def get_node_value(self, param_name: str) -> Any:
        # Get Node ID
        node = self.get_node(param_name)

        # Get Value
        return node.get_value()

    def set_node_value(self, param_name: str, value: Any) -> Any:
        # Get Node ID
        node = self.get_node(param_name)

        # Get Value
        return node.set_value(value)

    def make_step(self, is_last_step: bool = False):
        # Make step
        self._step_node.set_value(True)

        if not is_last_step:
            while self._step_node.get_value():
                pass

    def terminate(self):
        # Run the rest of the simulation (terminate command doesn't work)
        self._run_node.set_value(True)

    def get_simulation_time(self):
        return self._time.get_value()

class OpenModelicaSimulation(AbstractSimulation):

    # Default simulation values
    DefaultIsRealTime: bool = True
    DefaultPort: int = 4841
    DefaultSolver: str = "dassl"
    DefaultTolerance: float = 1e-12

    # Component name pattern
    ComponentPattern: str = "^(?P<component_name>\w+)\\.(?P<quantity_name>.+)$"
    
    # Solvers
    AvailableSolvers: list[str] = [
        "cvcode",
        "dassl",
        "euler", 
        "ida",
        "impeuleur",
        "imprungekutta",
        "irksco",
        "heun",
        "optimization",
        "qss",
        "rungekutta",
        "rungekuttaSsc",
        "symSolver",
        "symSolverSsc",
        "trapezoid"
    ]

    def __init__(self, om_file_path: str) -> None:
        super().__init__()

        # Change Directoryif not os.path.isabs(mo_filename):
        om_file_path = os.path.join(os.getcwd(), om_file_path).replace("\\","/")

        if os.path.exists(om_file_path):
            # Extract working directory
            self.working_directory = os.path.dirname(om_file_path)
            self.om_filename = os.path.basename(om_file_path)

            # Change current Directory before starting Open Modelica Processes
            DirectoryManager().set_working_directory(self.working_directory)

            # Init System
            self.om_system = ModelicaSystem(self.om_filename, ".".join(self.om_filename.split(".")[:-1]))

            if self.om_system.xmlFile is not None and self.om_system.tree is not None:
                # Get model informations
                self._inputs = self.om_system.getInputs()
                self._outputs = self.om_system.getOutputs()
                self._parameters = self.om_system.getParameters()
                self._components = self.get_components()

                # Init Simulation options with default values
                self._tolerance = OpenModelicaSimulation.DefaultTolerance
                self._solver = OpenModelicaSimulation.DefaultSolver

                # Port
                self._port = OpenModelicaSimulation.DefaultPort

                # Is real time simulation
                self._is_real_time = OpenModelicaSimulation.DefaultIsRealTime

                # Reset working directory
                DirectoryManager().reset_working_directory()

            else:
                # Reset working directory
                DirectoryManager().reset_working_directory()

                # Raise type error
                raise(TypeError(f"File is not a valid Open Modelica file : {self.om_system.xmlFile} {self.om_system.tree}"))

        else:
            raise(FileNotFoundError(f"Open Modelica file doesn't exists : {om_file_path}"))

    # --------------
    # Properties
    # --------------
    @property
    def tolerance(self) -> numbers.Number:
        return self._tolerance

    @tolerance.setter
    def tolerance(self, new_tolerance: numbers.Number) -> None:
        if OpenModelicaSimulation.is_valid_tolerance(new_tolerance) and new_tolerance != self._tolerance:
            self._tolerance = new_tolerance

    @property
    def solver(self) -> str:
        return self._solver

    @solver.setter
    def solver(self, new_solver: str) -> None:
        if OpenModelicaSimulation.is_valid_solver(new_solver) and new_solver != self._solver:
            self._solver = new_solver

    @property
    def port(self) -> numbers.Number:
        return self._port

    @port.setter
    def port(self, new_port: numbers.Number) -> None:
        if OpenModelicaSimulation.is_valid_port(new_port) and new_port != self._port:
            self._port = new_port

    @property
    def is_real_time(self) -> numbers.Number:
        return self._is_real_time

    @is_real_time.setter
    def is_real_time(self, new_is_real_time: numbers.Number) -> None:
        if OpenModelicaSimulation.is_valid_is_real_time(new_is_real_time) and new_is_real_time != self._is_real_time:
            self._is_real_time = new_is_real_time

    # --------------
    # Methods
    # --------------
    def get_components(self):
        components = {}
        for solution_name in self.om_system.getContinuous():
            continuous_item_value = self.om_system.getContinuous(solution_name)
            OpenModelicaSimulation.extract_continuous_information(components, solution_name, continuous_item_value)
        return components

    def init_simulation(self):
        # Set default value if unexpected values
        # Tolerance
        if not OpenModelicaSimulation.is_valid_tolerance(self._tolerance):
            self._tolerance = OpenModelicaSimulation.DefaultTolerance

        # Solver
        if not OpenModelicaSimulation.is_valid_solver(self._solver):
            self._solver = OpenModelicaSimulation.DefaultSolver

        # Init Simulation
        self.om_system.setSimulationOptions([
            f"startTime={self._start_time}",
            f"stopTime={self._stop_time}",
            f"stepSize={self._step_size}",
            f"tolerance={self._tolerance}",
            f"solver={self._solver}",
        ])

    def run_simulation(self):

        # Set Working Directory
        DirectoryManager().set_working_directory(self.working_directory)

        # Build Model
        self.om_system.buildModel()

        # Run Simulation
        simflags = [
            "-embeddedServer=opc-ua",
            f"-embeddedServerPort={self._port}"
        ]

        # Run Simulation
        try:
            if self._is_real_time:
                simflags.insert(0, "-rt=1.0")

            # Rn in a different thread
            self.simulation_thread = Thread(target=self.om_system.simulate, daemon=True, kwargs={"simflags": " ".join(simflags)})
            self.simulation_thread.start()
        except:
            pass

        # Set simulation controller
        self.simulation_controller = OpenModelicaSimulationController(f"opc.tcp://127.0.0.1:{self._port}")
        self.simulation_controller.init_connection()
        DirectoryManager().reset_working_directory()

    def set_simulation_options(
            self, 
            start_time: float = None, 
            stop_time: float = None, 
            step_size: float = None, 
            tolerance: float = None, 
            solver: str = None
        ):
        # Common parameters
        super().set_simulation_options(start_time, stop_time, step_size)

        # Tolerance
        if OpenModelicaSimulation.is_valid_tolerance(tolerance):
            self._tolerance = tolerance

        # Solver
        if OpenModelicaSimulation.is_valid_solver(solver):
            self._solver = solver

    def get_node_value(self, param_name: str) -> Any:
        return self.simulation_controller.get_node_value(param_name)

    def set_node_value(self, param_name: str, value: Any) -> Any:
        self.simulation_controller.set_node_value(param_name, value)

    def make_step(self, is_last_step: bool = False):
        self.simulation_controller.make_step(is_last_step)

    def terminate(self):
        # Terminate simulation
        self.simulation_controller.terminate()

        # Wait for thread finished
        self.simulation_thread.join()

        # Disconnect client
        self.simulation_controller.disconnect()

    def get_simulation_time(self):
        return self.simulation_controller.get_simulation_time()
    

    # --------------
    # Static Methods
    # --------------
    @staticmethod
    def extract_continuous_information(component_dict: dict, continuous_item: str, value, depth=1) -> dict:
        # Extract separate component & quantity name
        for match in re.finditer(OpenModelicaSimulation.ComponentPattern, continuous_item):
            # Get component name
            component_name = match.groupdict()["component_name"]

            # Get quantity name
            quantity_name = match.groupdict()["quantity_name"]

            # If component not already added → create it as dict
            if component_name not in component_dict:
                component_dict[component_name] = {}

            if re.match(OpenModelicaSimulation.ComponentPattern, quantity_name):
                # Handle sub quantity parameter
                sub_component_dict = component_dict[component_name]
                OpenModelicaSimulation.extract_continuous_information(sub_component_dict, quantity_name, value, depth+1)
            else:
                # Set value
                component_dict[component_name][quantity_name] = value

    @staticmethod
    def is_valid_tolerance(tolerance: Any):
        return tolerance is not None and isinstance(tolerance, numbers.Number)

    @staticmethod
    def is_valid_solver(solver: Any):
        return solver is not None and isinstance(solver, str) and solver in OpenModelicaSimulation.AvailableSolvers
    
    @staticmethod
    def is_valid_port(port: Any):
        return port is not None and isinstance(port, int)

    @staticmethod
    def is_valid_is_real_time(is_real_time: Any):
        return is_real_time is not None and isinstance(is_real_time, bool) 
    




# PyQt5 imports
from PyQt5.QtWidgets import QTabWidget

# Project imports
from muphyn.packages.interface.editors.abstract_editor import AbstractEditor
from muphyn.packages.interface.editors.diagram_editor import GraphicsView
from muphyn.packages.interface.graphical_actions.actions_holder import ActionsHolder
from muphyn.packages.interface.graphical_actions.diagram_add_graphical_element_action import DiagramAddGraphicalElementAction
from muphyn.packages.interface.graphical_actions.diagram_paste_graphical_element_action import DiagramPasteGraphicalElementAction
from muphyn.packages.interface.graphical_actions.diagram_remove_graphical_element_action import DiagramRemoveGraphicalElementAction
from muphyn.packages.interface.models.editable_models.abstract_diagram_model import AbstractDiagramModel
from muphyn.packages.interface.models.editable_models.abstract_editable_model import AbstractEditableModel
from muphyn.packages.interface.models.graphical_models.abstract_graphical_element import AbstractGraphicalElement
from muphyn.packages.interface.models.links_model.abstract_link_model import AbstractLinkModel

class MultiPhysicsDiagramEditor(AbstractEditor):

    def __init__(self, tab_holder: QTabWidget, diagram_model: AbstractDiagramModel, actions_holder: ActionsHolder):
        super().__init__(tab_holder, diagram_model, actions_holder)

        # Add graphics view
        self._graphics_view : GraphicsView = GraphicsView(self, self._actions_holder, diagram_model) 
        self._graphics_view.setGeometry(0, 0, self.width(), self.height())

        # Activate mouse tracking
        self.setMouseTracking(True)

    # -------------
    # Properties
    # -------------
    @property
    def diagram_model(self) -> AbstractDiagramModel:
        return self._editable_model

    # -------------
    # Core Methods
    # -------------
    def add_graphical_element (self, graphical_element : AbstractGraphicalElement) -> None :
        """Permet d'ajouter un élément graphique à l'interface."""

        if graphical_element is None :
            return

        self.diagram_model.add_element(graphical_element)

    def rem_graphical_element (self, graphical_element : AbstractGraphicalElement) -> None :
        """Permet de supprimer un élément graphique de l'interface."""

        self.diagram_model.remove_element(graphical_element)
        graphical_element.deleteLater()

    def add_item (self, graphical_element : AbstractGraphicalElement) -> None :
        self._graphics_view.add_graphical_element(graphical_element)
        
    def rem_item (self, graphical_element : AbstractGraphicalElement) -> None :
        self._graphics_view.rem_graphical_element(graphical_element)

    def clear (self) -> None :
        to_delete = [el for el in self.elements() if not isinstance(el, AbstractLinkModel)]
        for el in to_delete :
            self.rem_item(el)

    # -------------
    # Graphical Methods
    # -------------
    def delete_selection (self) -> None :
        action = DiagramRemoveGraphicalElementAction(self._graphics_view._diagram_model, self.selected_elements())
        action.do()
        self.actions_holder.append(action)

class OpenModelicaDiagramEditor(MultiPhysicsDiagramEditor):

    def __init__(self, tab_holder: QTabWidget, diagram_model: AbstractDiagramModel, actions_holder: ActionsHolder):
        super().__init__(tab_holder, diagram_model, actions_holder)



from muphyn.packages.core.plci_core_diagram import Diagram
from muphyn.packages.core.plci_core_scheduler import Scheduler
from muphyn.packages.core.plci_core_scheduler_params import SchedulerParams
from .unit_test import get_boxes_libraries, get_schedulers_library, simulate

# Get library
boxesLibraries = get_boxes_libraries()
schedulersLibraries = get_schedulers_library()

totalLibraryCount : int = 0

for box in boxesLibraries.boxes:
    print('Box loaded : ', box.box_library, '.', box.box_name)
    totalLibraryCount += 1

print("Nombre de boxes importées : ", totalLibraryCount)
print("13 boxes en code")
print("3 boxes composites (dont 1 dépendante de l'autre)")

diagram : Diagram = boxesLibraries.construct_box('Boxes.User', 'Wrapper')

print("Diagram loaded : ", diagram)
print("Diagram boxes : ", diagram.boxes.__len__())
print("Diagram signals : ", diagram.signals.__len__())
print("Diagram box input : ", diagram._box_inputs.__len__())
print("Diagram linked signals : ", diagram._linked_signals.__len__())

scheduler : Scheduler = schedulersLibraries.construct_scheduler('Schedulers', 'Default')
scheduler.params = SchedulerParams(10, 0.01)
scheduler.diagram = diagram

simulate(scheduler)


import typing

from PyQt5.QtCore import QPoint, QPointF
from PyQt5.QtWidgets import QApplication, QMainWindow, QGraphicsObject

def findMainWindow() -> typing.Union[QMainWindow, None]:
    # Global function to find the (open) QMainWindow in application
    app = QApplication.instance()
    for widget in app.topLevelWidgets():
        if isinstance(widget, QMainWindow):
            return widget
    return None


#-----------------------------------
# Imports
#-----------------------------------

import sys
from typing import Callable, Iterator, List, Any, Dict
from muphyn.packages.core.plci_core_data import Data

from muphyn.packages.core.plci_core_data_type import DataType
from muphyn.packages.core.plci_core_scheduler_params import SchedulerParams
from muphyn.packages.core.plci_core_signal_event import SignalEvent
from muphyn.packages.core.plci_core_scheduler_event import SchedulerEvent
from muphyn.packages.core.plci_core_signal import Signal

#-----------------------------------
# Class
#-----------------------------------

class Box :
    """Est la classe décrivant les boxes."""
    # -------------
    # Constructors
    # -------------

    def __init__ (  self, 
                    index_ : int, 
                    name_ : str,
                    library_ : str,
                    init_method_ : Callable[[Any], None] = None, 
                    function_ : Callable[[Any, SchedulerEvent], List[SignalEvent]] = None, 
                    end_method_ : Callable[[Any], None] = None, 
                    wait_for_events_ : bool = True,
                    wait_for_all_signal_events_ : bool = True, 
                    params_ : Dict[str, Any] = None) :
        
        self._index : int = index_
        self._name : str = name_
        self._library : str = library_
        
        if params_ is None :
            self._params : Dict[str, Any] = {}
        else :
            self._params : Dict[str, Any] = params_

        self._inputs : Dict[Signal, float] = {}
        self._outputs : List[Signal] = []

        self._wait_for_events : bool = wait_for_events_
        self._wait_for_all_signal_events : bool = wait_for_all_signal_events_

        self._current_timing : float = 0.0

        if (init_method_ == None) : 
            self._init_method : Callable[[Box], None] = lambda box, scheduler_params : ...
        else :
            self._init_method : Callable[[Box], None] = init_method_

        if (function_ == None) :
            self._function : Callable[[Box, SchedulerEvent], List[SignalEvent]] = lambda box, timing : []
        else :
            self._function : Callable[[Box, SchedulerEvent], List[SignalEvent]] = function_

        if (end_method_ == None) :
            self._end_method : Callable[[Box], None] = lambda box : ...
        else :
            self._end_method : Callable[[Box], None] = end_method_
        
    # -------------
    # Properties
    # -------------

    @property 
    def index (self) -> int :
        """Permet de récuperer l'index de la box."""
        return self._index

    @property
    def name (self) -> str :
        """Permet de récuperer le nom de la box."""
        return self._name

    @property
    def library (self) -> str :
        """Permet de récuperer la bibliothèque dans laquelle se trouve la box."""
        return self._library
    
    @property
    def inputs (self) -> List[Signal] :
        """Permet de récuperer les entrées de la box."""
        return self._inputs.keys()

    @property
    def outputs (self) -> List[Signal] :
        """Permet de récuperer les sorties de la box."""
        return self._outputs

    @property
    def params (self) -> Dict[str, Any] :
        """Permet de récuperer les paramètres de le box."""
        return self._params

    @property
    def wait_for_events (self) -> bool :
        """Permet de savoir si la box ne peut être utilisée que lors d'un événement."""
        return self._wait_for_events

    @property
    def wait_for_all_signal_events (self) -> bool :
        """Permet de savoir si la box ne peut être appelée que lorsque toutes ses entrées ont reçues un événement au timing actuel."""
        return self._wait_for_all_signal_events

    @property 
    def events_on_all_inputs (self) -> bool :
        """Permet de savoir si toutes les entrées ont bien reçues un événement au timing actuel."""
        for signal in self._inputs.keys() :
            if self._inputs[signal] < self.current_timing : 
                return False
        return True

    @property 
    def current_timing (self) -> float :
        """Permet de récuperer le timing de la box."""
        return self._current_timing


    # -------------
    # Methods
    # -------------

    def init (self, scheduler_params: SchedulerParams) -> None :
        """Permet de lancer la méthode d'initialisation de la box."""
        self._init_method(self, scheduler_params)

    def function (self, event_ : SchedulerEvent) -> List[SignalEvent] :
        """Permet de réaliser le traitement de la box et de modifier les éventuels sorties."""

        self._current_timing = event_.timing
        
        # Si on attend que des événements et que ce n'est pas un événement. On s'arrête
        if self.wait_for_events :
            if event_.signal is None :
                return []
            
        # Si c'est bien un événement, on modifie le timing de l'entrée concernée.
        if not event_.signal is None :
            self._inputs[event_.signal] = event_.timing
        
        # Si on attends tous les événement et qu'un événement n'est pas encore arrivé. On s'arrête.
        if self.wait_for_all_signal_events :
            if not self.events_on_all_inputs :
                return []

        # Activation de la box.
        return_data = self._function(self, event_)

        # Si l'activation n'a rien retourné, on produit quand même un tableau vide.
        if return_data is None :
            return_data : List[SchedulerEvent] = []

        # Retour du tableau d'activation.
        return return_data

    def end (self) -> None :
        """Permet d'appeler la méthode de fin de la box une fois que le planificateur a terminé."""
        self._end_method(self)

    def append_outputs (self, *outputs : Signal) -> None :
        """Permet d'ajouter autant de sorties que voulue à la box."""
        for output in outputs :
            self._outputs.append(output)

    def append_inputs (self, *inputs : Signal) -> None : 
        """Permet d'ajouter autant d'entrée que voulue à la box."""
        for input in inputs :
            self._inputs[input] = -1

    def get_input (self, input_index : int) -> Signal :
        """Permet de récuperer un index suivant son index dans la liste de la boite actuelle."""
        for input in self._inputs :
            if input_index == 0 :
                return input
            
            input_index -= 1

        return None

    def get_parameter(self, name: str):
        if name in self._params:
            return self._params[name]
        else:
            raise(Exception(f"No {name} parameter in Box parameters -> {[param for param in self._params]}"))

    def construct_signal_event (self, output : Signal, value : Any) -> SignalEvent :
        """Permet de générer des events signal."""
        

        if isinstance(value, float) :
            if output.signal_type == DataType.FLOAT :
                data = Data(DataType.FLOAT, value)

            elif output.signal_type == DataType.INT :
                data = Data(DataType.INT, int(value))

        elif isinstance(value, str) :
            data = Data(DataType.STRING, value)
        
        elif isinstance(value, int) :
            if output.signal_type == DataType.FLOAT :
                data = Data(DataType.FLOAT, float(value))

            elif output.signal_type == DataType.INT :
                data = Data(DataType.INT, value)

        elif isinstance(value, object) :
            data = Data(DataType.OBJECT, value)

        else :
            data = Data(DataType.UNDIFINED, value)
        
        return SignalEvent(output, self, data)

    def __str__ (self) :
        """Permet de retourner la box sous la forme d'un string."""
        self._params.__setitem__
        return "Box [" + self._library + " " + self._name + " | " + str(self._index) + "]"

    def __getitem__ (self, name : str) -> Any :
        """Permet de récuperer un objet des paramètres en utilisant les crochets."""

        if name in self._params :
            return self._params.__getitem__(name)

        else :
            return None

    def __setitem__ (self, name : str, obj : Any) -> None :
        """Permet de modifier un paramètre en utilisant les crochets."""
        self._params.__setitem__(name, obj) 
        self._params.__iter__()

    def __delitem__ (self, name : str) -> None :
        """Permet de supprimer un paramètre en utilisant son nom."""
        self._params.__delitem__(name)

    def __iter__ (self) -> Iterator[str] :
        """Permet d'itérer dans les paramètres."""
        return self._params.__iter__()

    if sys.version_info >= (3, 8):
        def __reversed__ (self) -> Iterator[str] :
            """Permet de récuperer la version inversée de la liste des paramètres."""
            return self._params.__reversed__()

    def __contains__ (self, obj : str) -> bool : 
        """Permet de savoir si l'a box cotient l'élément passé en in."""
        return self._params.__contains__(obj)



#-----------------------------------
# Imports
#-----------------------------------

import traceback
from types import TracebackType
from typing import List

from muphyn.packages.core.utils.log_manager import LogManager
from .plci_core_diagram import Diagram
from .plci_core_box import Box
from .plci_core_signal_event import SignalEvent

#-----------------------------------
# Class
#-----------------------------------

class SchedulerException :
    """Est la classe qui permet de créer un retour lors d'une exception dans un planificateur."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, box_ : Box, box_bis_ : Box, events_ : List[SignalEvent], event_ : SignalEvent, diagram_ : Diagram, timing_ : float, exception_ : Exception) :
        self._box = box_
        self._box_bis = box_bis_
        self._events = events_
        self._event = event_
        self._diagram = diagram_
        self._exception = exception_
        self._timing = timing_

    
    # -------------
    # Methods
    # -------------

    def print (self) :
        to_print_rows = []
        to_print_rows.append("SCHEDULER EXCEPTION : ")

        to_print_rows.append(f"\tException at : {self._timing: .3f}s")

        if self._box is None :
            to_print_rows.append("\tBox : No current box")
        else :
            to_print_rows.append(f"\tBox : {self._box.library} {self._box.name} | index : {self._box.index}")

        if self._box_bis is None :
            to_print_rows.append("\tBis box : No current bis box")
        else :
            to_print_rows.append(f"\tBix Box : {self._box.library} {self._box.name} | index : {self._box_bis.index}")

        if self._event is None :
            to_print_rows.append("\tCurrent event : No current event")
        else :
            to_print_rows.append(f"\tCurrent event : box index : {self._event.box.index} | signal index : {self._event.signal.index}")

            if not self._event.signal in self._diagram.box_inputs :
                to_print_rows.append("\tThe signal does not have any box to tickle !!!")

        if self._events is None :
            to_print_rows.append("\tEvents list : No events list")
        else :
            to_print_rows.append(f"\tEvents list : {len(self._events)} events in the queue")

        to_print_rows.append(f"\t{''.join(traceback.format_exception(self._exception))}")

        LogManager().error('\n'.join(to_print_rows))


#-----------------------------------
# Imports
#-----------------------------------

# General Imports
import copy
from typing import Any, Dict

# PyQt5 Imports
from PyQt5.QtGui import QColor

# Project Imports
from muphyn.packages.interface.models.graphical_models.abstract_box_model import AbstractBoxModel
from muphyn.packages.interface.models.graphical_models.abstract_graphical_element import AbstractGraphicalElement
from muphyn.packages.interface.models.graphical_models.box_input_model import BoxInputModel
from muphyn.packages.interface.models.graphical_models.box_model import BoxModel
from muphyn.packages.interface.models.graphical_models.box_output_model import BoxOutputModel
from muphyn.packages.interface.models.links_model.abstract_link_model import AbstractLinkModel
from muphyn.packages.interface.models.signals_model.input_connection_model import InputConnectionGroupModel, InputConnectionModel
from muphyn.packages.interface.models.signals_model.output_connection_model import OutputConnectionGroupModel, OutputConnectionModel

#-----------------------------------
# Functions
#-----------------------------------
def link (link : AbstractLinkModel) -> None :
    """Permet de peupler la liste de reconstructeur avec les éléments nécessaire pour reconstruire le lien actuel."""

    reconstructor = {}
    reconstructor["type"] = 'link'

    reconstructor["graphical_index"] = link.graphical_index
    reconstructor["input_box_index"] = link.input.box_model.graphical_index
    reconstructor["output_box_index"] = link.output.box_model.graphical_index
    reconstructor["input_index"] = link.input.graphical_index
    reconstructor["output_index"] = link.output.graphical_index

    reconstructor["link_type"] = link.link_type
    reconstructor["link_value"] = link.link_value
    reconstructor["link_text"] = link.text

    return copy.deepcopy(reconstructor)

def box (box : BoxModel) -> Dict : 
    """Permet de peupler la liste de reconstructeur avec les éléments nécessaire pour reconstruire une box."""

    # Common box informations
    reconstructor = abstract_box(box)

    # General box informations
    reconstructor["type"] = 'box'
    reconstructor["library"] = box.library

    # Parameters
    reconstructor["params"] = {}
    for param_name in box.get_parameters() :
        reconstructor["params"][param_name] = {}
        reconstructor["params"][param_name]["type"] = box.get_parameter(param_name)["type"]
        reconstructor["params"][param_name]["value"] = box.get_parameter(param_name)["value"]
    
    return copy.deepcopy(reconstructor)

def composite_box_input (input : BoxInputModel) -> Dict :
    """Permet de peupler la liste de reconstructeur avec les éléments nécesaire pour reconstruire une entrée de box composite."""

    reconstructor = abstract_box(input)
    reconstructor["type"] = 'box-composite-input'
    
    return copy.deepcopy(reconstructor)

def composite_box_output (output : BoxOutputModel) -> Dict :
    """Permet de peupler la liste de reconstructeur avec les éléments nécessaire pour reconstruire une sortie de box composite."""

    reconstructor = abstract_box(output)
    reconstructor["type"] = 'box-composite-output'
    
    return copy.deepcopy(reconstructor)

def box_inputs_group(inputs_group: InputConnectionGroupModel, inputs_group_reconstructor : Dict = {}) -> dict:
    # Inputs group informations
    inputs_group_reconstructor["name"] = inputs_group.name
    inputs_group_reconstructor["type"] = inputs_group.data_type
    inputs_group_reconstructor["isInfinite"] = inputs_group.is_infinite

    # Inputs informations
    inputs_group_reconstructor["inputs"] = [box_input(input_) for input_ in inputs_group.inputs]

    return copy.deepcopy(inputs_group_reconstructor)

def box_input (input_ : InputConnectionModel, input_reconstructor : Dict = {}) -> Dict :
    """Permet de créer un dictionnaire contenant les données d'une entrée de box."""
    input_reconstructor["graphical_index"] = input_.graphical_index
    input_reconstructor["name"] = input_.name
    input_reconstructor["type"] = input_.data_type
    input_reconstructor["isInfinite"] = input_.is_infinite
    input_reconstructor["text"] = input_.text
    # input_reconstructor["index"] = input_.parent().index_of_input(input_)

    return copy.deepcopy(input_reconstructor)

def box_outputs_group(outputs_group: OutputConnectionGroupModel, outputs_group_reconstructor : Dict = {}) -> dict:
    # Outputs group informations
    outputs_group_reconstructor["name"] = outputs_group.name
    outputs_group_reconstructor["type"] = outputs_group.data_type
    outputs_group_reconstructor["isInfinite"] = outputs_group.is_infinite

    # Outputs informations
    outputs_group_reconstructor["outputs"] = [box_output(output_) for output_ in outputs_group.outputs]

    return copy.deepcopy(outputs_group_reconstructor)

def box_output (output : OutputConnectionModel, output_reconstructor : Dict = {}) -> Dict :
    """Permet de créer un dictionnaire contenant les données d'une soirte de box."""
    
    output_reconstructor["graphical_index"] = output.graphical_index
    output_reconstructor["name"] = output.name
    output_reconstructor["type"] = output.data_type
    output_reconstructor["isInfinite"] = output.is_infinite
    output_reconstructor["text"] = output.text
    # output_reconstructor["index"] = output.parent().index_of_output(output)

    return copy.deepcopy(output_reconstructor)

def abstract_box (abstract_box : AbstractBoxModel) -> Dict :
    """Permet de peupler un reconstructeur avec les éléments d'une box abstraite."""

    reconstructor = abstract_graphical_element(abstract_box)

    # Inputs
    reconstructor["inputs_groups"] = [box_inputs_group(inputs_group) for inputs_group in abstract_box.inputs_groups.values()]

    # Outputs
    reconstructor["outputs_groups"] = [box_outputs_group(outputs_group) for outputs_group in abstract_box.outputs_groups.values()]
    reconstructor["outputs"] = []
    for output in abstract_box.outputs : 
        reconstructor["outputs"].append(box_output(output))

    return copy.deepcopy(reconstructor)

def abstract_graphical_element (abstract_graphical_element : AbstractGraphicalElement) -> Dict :
    """Permet de peupler un reconstructeur avec les éléments d'un graphical element."""

    reconstructor = {}

    reconstructor["graphical_index"] = abstract_graphical_element.graphical_index
    reconstructor["geometry"] = {}
    reconstructor["geometry"]["x"] = abstract_graphical_element.position.x()
    reconstructor["geometry"]["y"] = abstract_graphical_element.position.y()
    reconstructor["geometry"]["width"] = abstract_graphical_element.size.width()
    reconstructor["geometry"]["height"] = abstract_graphical_element.size.height()
    reconstructor["geometry"]["rotation"] = abstract_graphical_element.rot
    reconstructor["name"] = abstract_graphical_element.name
    reconstructor["text"] = abstract_graphical_element.text

    return copy.deepcopy(reconstructor)
    

def encode (element : Any) -> Dict :
    """Permet de sélectionne la bonne méthode à appeler pour peupler le reconstructeur."""

    if isinstance(element, AbstractLinkModel) :
        return link(element)

    elif isinstance(element, BoxModel) :
        return box(element)

    elif isinstance(element, BoxInputModel) :
        return composite_box_input(element)

    elif isinstance(element, BoxOutputModel) :
        return composite_box_output(element)



from muphyn.packages.core.box_library.plci_core_boxes_libraries import BoxesLibraries
from muphyn.packages.core.plci_core_scheduler import Scheduler
from muphyn.packages.core.plci_core_data_type import DataType
from muphyn.packages.core.plci_core_signal import Signal
from muphyn.packages.core.plci_core_scheduler_params import SchedulerParams
from muphyn.packages.core.scheduler_library.plci_core_schedulers_libraries import SchedulersLibraries

from muphyn.packages.unit_test.unit_test import get_boxes_libraries, get_schedulers_library, simulate

# Get library
boxesLibraries : BoxesLibraries = get_boxes_libraries()
schedulersLibraries : SchedulersLibraries = get_schedulers_library()

# Parameters
scheduler : Scheduler = schedulersLibraries.construct_scheduler('Schedulers', 'Default')
scheduler.params = SchedulerParams(stop_time_ = 0.2, step_time_ = 0.0001)


# Signal de source (Carré)
s0 = Signal(0, DataType.FLOAT, 0.0)
scheduler.diagram.append(s0)

# Signal dérivé
s1 = Signal(1, DataType.FLOAT, 0.0)
scheduler.diagram.append(s1)

# Sources (Carré)
period = scheduler.stop_time / 3
high_time = period / 3
print("Square wave period : ", period)
print("High time : ", high_time)
b0 = boxesLibraries.construct_box('Boxes.Sources', 'Square', high_value = 1, low_value = 0, period = period, high_time = high_time)
scheduler.diagram.append(b0)

# Boxe de dérivation
b1 = boxesLibraries.construct_box('Boxes.Math', 'Derivator')
scheduler.diagram.append(b1)

# Sink
b2 = boxesLibraries.construct_box('Boxes.Sinks', 'Graph', start_time = 0, stop_time = scheduler.stop_time, title = "square waves", label_x = "Temps (s)", label_y = "Amplitude (v)")
scheduler.diagram.append(b2)

b3 = boxesLibraries.construct_box('Boxes.Sinks', 'Graph', start_time = 0, stop_time = scheduler.stop_time, title = "Derivative signal", label_x = "Temps (s)", label_y = "Amplitude (v)")
scheduler.diagram.append(b3)

# Diagram creation
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[0], scheduler.diagram.signals[0])
scheduler.diagram.add_box_outputs(scheduler.diagram.boxes[1], scheduler.diagram.signals[1])

scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[1], scheduler.diagram.signals[0])
scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[2], scheduler.diagram.signals[0])
scheduler.diagram.add_box_inputs(scheduler.diagram.boxes[3], scheduler.diagram.signals[1])


# Launch solver
simulate(scheduler)


# -----------------------------------
# Imports
# -----------------------------------

from typing import List
from scipy.integrate import odeint

# -----------------------------------
# Functions
# -----------------------------------

def get_coeff_vector(coeff_vector_string: str) -> List[float]:

    if coeff_vector_string is None:
        return [0]

    vector = []
    for sub in coeff_vector_string.split(' '):

        sub = sub.strip()
        if len(sub) == 0:
            continue

        try:
            vector.append(float(sub))
        except ValueError:
            ...

    return vector


def _init_transfer_function(box, simulation_params) -> None:
    
    box['num_vect'] = get_coeff_vector(box['numerator'])
    box['num_vect'].reverse()
    box['num_order'] = len(box['num_vect']) - 1

    box['denom_vect'] = get_coeff_vector(box['denominator'])
    box['denom_vect'].reverse()
    box['denom_order'] = len(box['denom_vect']) - 1

    box['last_y'] = [box['initial_value']] + [0] * (box['denom_order'] -1) 
    
    box['last_timing'] = -1

def f(y, t, box): 
    out = 0
    for j, coeff_y  in enumerate(box['denom_vect'][:-1]):
        out -= coeff_y * y[j]
    
    for j, coeff_y in enumerate(box['num_vect']):
        if j == 0:
            out += coeff_y * box.get_input(0).value
        else:
            out += coeff_y * y[j]
    
    out_vector = [y[box['denom_order'] - 1 - x] for x in range(box['denom_order'])]
    out_vector[-1] = out/box['denom_vect'][-1]

    return out_vector

def _function_transfer_function(box, event_) -> List :             

    if box['last_timing'] == event_.timing:    # if we are still on the current step, leave
        return []

    events: List = []

    if box['last_timing'] == -1:
        box['last_timing'] = event_.timing

        for output in box.outputs :
            events.append(box.construct_signal_event(output, box['initial_value']))
    else:
        step = event_.step_time

        t0 = event_.timing

        t = [t0, t0 + step]                 # time vector is only current and next step

        sol = odeint(f, box['last_y'], t, args=(box,))   # run odeint until next step
        
        out = list(sol[-1,:])               # use the last values as the output of the step

        box['last_y'] = out

        for output in box.outputs :
            events.append(box.construct_signal_event(output, out[0])) # only send the y(t) output, discard eventual derivatives for order >= 2
    
    box['last_timing'] = event_.timing

    return events



#-----------------------------------
# Imports
#-----------------------------------

from typing import Any

from muphyn.packages.interface.graphical_actions.abstract_diagram_action import AbstractDiagramAction
from muphyn.packages.interface.models.links_model.abstract_node_model import AbstractNodeModel
from muphyn.packages.interface.models.signals_model.abstract_signal_model import AbstractSignalModel
from muphyn.packages.interface.models.signals_model.signal_link_model import SignalLinkModel

#-----------------------------------
# Class
#-----------------------------------

class DiagramLinkNodesAction (AbstractDiagramAction) :
    """Est le type d'action qui permet de lier des noeuds entre eux."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, input_node : AbstractNodeModel, output_node : AbstractNodeModel) :

        AbstractDiagramAction.__init__(self, input_node.diagram_model)
        
        self._input_box_index = input_node.box_model.graphical_index
        self._output_box_index = output_node.box_model.graphical_index
        self._input_index = input_node.graphical_index
        self._output_index = output_node.graphical_index

        self._link_index = -1

    # -------------
    # Methods
    # -------------

    def get_input (self) -> Any :
        """Permet de récuperer l'entrée à lier."""

        box_element = self.diagram_model.get_element_by_graphical_index(self._input_box_index)

        if box_element is None : 
            return None 

        if hasattr(box_element, 'inputs_groups'):
            for inputs_group in box_element.inputs_groups.values():
                for input_ in inputs_group.inputs:
                    if input_.graphical_index == self._input_index:
                        return input_

        return None

    def get_output (self) -> Any :
        """Permet de récuperr la sortie à lier."""

        box_element = self.diagram_model.get_element_by_graphical_index(self._output_box_index)

        if box_element is None : 
            return None 

        if hasattr(box_element, 'outputs') :
            for output in box_element.outputs :
                if output.graphical_index == self._output_index :
                    return output

        return None

    def do (self) :
        input_ = self.get_input()
        output = self.get_output()

        # Set input_ as connected
        input_.is_connected = True

        link = self.diagram_model.link_nodes(input_, output)

        if self._link_index == -1 :
            self._link_index = link.graphical_index
        
        else :
            link.graphical_index = self._link_index

    def undo (self) :
        link : SignalLinkModel = self.diagram_model.get_element_by_graphical_index(self._link_index)
        if not(link is None) :
            
            # Get Input/Output BoxModel Graphical Index
            self._input_box_index = link.input.box_model.graphical_index
            self._output_box_index = link.output.box_model.graphical_index

            # Get Input/Output Graphical Index
            self._input_index = link.input.graphical_index
            self._output_index = link.output.graphical_index

            link.unbind()
            self.diagram_model.remove_element(link)
